from PyQt5 import QtCore, QtGui, QtWidgets
import sys


# ------------------------------------------------------------------------------
#           ArgolGSE_QtUI_Application
# ------------------------------------------------------------------------------
class ArgolGSE_QtUI_Application(QtWidgets.QApplication):
    """
    Starts up the QT Tabbed dialog box GUI used by the Argol GSE library. This class is a child of the
    QT application. The code has been modified so it can be used on several GSE projects.
    """

    def __init__(self, qt_userinterface_class):
        """
        The constructor loads up the tabbed widger defined in the *.ui file generated
        by QT Designer for this GSE project. The code looks for the user interface defined by method Ui_TabWidget()

        :param qt_userinterface_module:
        The QT GUI ui module generated from the QT Designer ui file.
        The python file is made by calling somthing like:
        pyrcc5 -o argolgse_resources_rc.py argolgse_resources.qrc
        pyuic5 --import-from=. -x -o argolgse_tabs.py  argolgse_tabs.ui
        """

        # self.setAttribute(QtCore.Qt.AA_DisableHighDpiScaling)
        # self.setAttribute(QtCore.Qt.AA_Use96Dpi)
        super().__init__(sys.argv)
        # self.setAttribute(QtCore.Qt.AA_DisableHighDpiScaling)
        self.setAttribute(QtCore.Qt.AA_Use96Dpi)
        # self.setAttribute(QtCore.Qt.AA_DontUseNativeDialogs)
        self.tab_widget = QtWidgets.QTabWidget()  # Create a QT tabWidget object
        self.ui = qt_userinterface_class()  # Now create the python interface to the GUI interface
        self.ui.setupUi(self.tab_widget)  # Populate the tab widget
        self.tab_widget.show()  # and show the object

    # ------------------------------------------------------------------------------
    #           notify
    # ------------------------------------------------------------------------------
    def notify(self, receiver: QtCore.QObject, event: QtCore.QEvent):
        """
        Implements the event notify that includes checking for the User defined events that come from
        the GSE code when it is asynchronously reading the HTML code from the 
        :param receiver: 
        :param event: 
        :return: 
        """
        if event.type() > QtCore.QEvent.User:
            w = receiver
            while (w):
                # Note that this calls `event` method directly thus bypassing
                # calling qApplications and receivers event filters
                res = w.event(event)
                if res and event.isAccepted():
                    return res
                w = w.parent()
        return super().notify(receiver, event)


# ------------------------------------------------------------------------------
#           my_exception_hook
# ------------------------------------------------------------------------------
sys._excepthook = sys.excepthook


def my_exception_hook(exctype, value, traceback):
    # Print the error and traceback
    print(exctype, value, traceback)
    # Call the normal Exception hook after
    sys._excepthook(exctype, value, traceback)
    sys.exit(1)


# Set the exception hook to our wrapping function
sys.excepthook = my_exception_hook


# ------------------------------------------------------------------------------
#           run_guiapplication
# ------------------------------------------------------------------------------

def run_guiapplication(instrument_globalobject_class, qt_userinterface_class):
    """
    Runs the GUI application. This will load the QT designer UI Tabbed dialog interface  which has been compiled
    and exposrts an instance of `qt_userinterface_class`. This runs the application, initializes the main global object
    and listens for events from the main application until an exit is signaled.

    :param instrument_globalobject_class:
        The instrument module that contains instrument specific code. Typically the mainglobalobject from `catsgse`, `aligse` or `lifegse`.
        This module has been imported by the caller.

    :param qt_userinterface_class:
        The python module from the QT GUI UI file  with no extension and no folder structure. This module has been imported by the caller

    """

    qtapp = ArgolGSE_QtUI_Application(qt_userinterface_class)       # Get the main QT application up and running
    mainglobalobject = instrument_globalobject_class(qtapp)         # Create the Main GSE global object. Internally loads QT resources and icons etc from the QT GUI UI module.
    qtapp.exec_()


#   sys.exit(qtapp.exec_())                                                                                # Run the loop


# -----------------------------------------------------------------------------------------------------------------------
#           class ArgolGSE_MainGlobalObject
# -----------------------------------------------------------------------------------------------------------------------

class ArgolGSE_MainGlobalObject:
    """
    A base class for the main GSE object. The main object holds the QT user interface and creates all of the widgets
    """

    def __init__(self, qtapp: ArgolGSE_QtUI_Application):
        self._gui_app = qtapp
#        self._sensorvariable_widgets = []

    # ------------------------------------------------------------------------------
    #           Property gui_app
    #   The gui application can be read but cannot be reset after initialization.
    # ------------------------------------------------------------------------------

    @property
    def gui_app(self):
        return self._gui_app
