from typing import Dict, Any, List, Callable
import datetime
import logging
from enum import IntEnum
import threading
from PyQt5 import QtCore, QtWidgets
from .network.udp_listener import UDP_Listener
from .network.udp_transmit import UDP_Transmit
from .yaml_configuration import Yaml_Configuration


# -----------------------------------------------------------------------------------------------------------------------
#           TrafficLight
#   A simple enumeration that maps a traffic light color to a display icon.
# -----------------------------------------------------------------------------------------------------------------------

class CommandStatusLight(IntEnum):
    """
    A small enumeration to hold  the state of the GUI of the traffic light statuses.
    Some code may rely upon the ordering so dont change the numerical ordering
    """

    Good = 0
    Waiting = 1
    Bad = 2


# ------------------------------------------------------------------------------
#           Command_History_Entry
# ------------------------------------------------------------------------------

class Command_History_Entry:

    def __init__(self, index: int, cmd: str):
        self.utc: datetime.datetime = datetime.datetime.utcnow()
        self.command: str = cmd
        self.reply: str = ""
        self.index: int = index
        self.tableview_row: int = -1
        self.receive_status: CommandStatusLight = CommandStatusLight.Waiting
        self.command_status: CommandStatusLight = CommandStatusLight.Waiting


# ------------------------------------------------------------------------------
#           CMD_Listener
# ------------------------------------------------------------------------------

class CMD_Listener:

    def __init__(self, qtapp: QtWidgets.QApplication,
                 udpconfig: Dict[str, Any],
                 config: Dict[str, Any]):
        #       config = { 'udp_on'  : True,
        #                  'name'    : 'CMDECHO',
        #                  'udp_ip'  : '10.65.0.33',
        #                  'udp_port': 50101,
        #                  'udp_max_message' : 2000,
        #                  'timeout' : -1
        #                  }

        self.udp = UDP_Listener(qtapp, udpconfig, config)

    # ------------------------------------------------------------------------------
    #           hk_callback
    # ------------------------------------------------------------------------------

    def cmd_callback(self, cmd_packet):
        print("received CMD ECHO packet: ", cmd_packet)


# ------------------------------------------------------------------------------
#           CMD_Transmit
# ------------------------------------------------------------------------------

class CMD_Transmit:

    def __init__(self, networkconfig: Dict[str, Any], config: Dict[str, Any]):
        self.udp = UDP_Transmit(networkconfig, config)

    # ------------------------------------------------------------------------------
    #           send
    # ------------------------------------------------------------------------------

    def send(self, cmd_packet):
        self.udp.send(cmd_packet)

    def local_address(self):
        return self.udp.udp_ipaddress


# -----------------------------------------------------------------------------------------------------------------------
#           class CMD_Manager
# -----------------------------------------------------------------------------------------------------------------------

class CMD_Manager:
    """
    A class that manges sending commands to the instrument
    """
    def __init__(self, qtapp: QtWidgets.QApplication, config_folder=None):
        super().__init__()

        config = Yaml_Configuration('network.yaml', basedir=config_folder)

        self.command_history: Dict[int, Command_History_Entry] = {}                                                             # The complete list of commands sent
        self.cmd_echo: CMD_Listener = CMD_Listener(qtapp, config.configuration["Network"], config.configuration["CMDECHO"])     # Listens to the UDP port for messages coming back from command system
        self.cmd_send: CMD_Transmit = CMD_Transmit(config.configuration["Network"], config.configuration["CMDSEND"])            # sends commands to the UDP port
        self.next_command_id: int = 100                                                                                         # The next command id to use
        self.callbacks: List[Callable] = []
        self.customcallback: Dict[int, Callable] = {}
        self.pending_recv_ack: List[int] = []                                                                                   # The list of commands receives we are waiting for
        self.pending_cmd_ack: List[int] = []                                                                                    # The list of commands executed  we are waiting for.
        self.lock: threading.Lock = threading.Lock()
        self.cmd_echo.udp.add_listener(self.echo_callback)                                                                      # Echo receive messages to this thread, using the event mapper of the main application

    # ------------------------------------------------------------------------------
    #           add_callback
    # ------------------------------------------------------------------------------

    def add_callback(self, callback_func: Callable):

        self.callbacks.append(callback_func)

    # ------------------------------------------------------------------------------
    #           send_command
    # ------------------------------------------------------------------------------

    def send_command(self, command: str, customcallback: Callable = None) -> int:

        id = self.next_command_id
        fullmessage = "%d %s" % (id, command)
        self.pending_recv_ack.append(self.next_command_id)
        self.pending_cmd_ack.append(self.next_command_id)
        if customcallback is not None:
            self.customcallback[self.next_command_id] = customcallback
        self.next_command_id += 1
        self.cmd_send.send(fullmessage)
        self.command_history[id] = Command_History_Entry(id, command)
        return id

    # ------------------------------------------------------------------------------
    #           echo_callback
    # ------------------------------------------------------------------------------

    def echo_callback(self, message: str):

        self.lock.acquire()

        if (message is not None) and (
                len(message) > 0):  # None or blank occurs if there was an error (eg timeout receiving the commandss
            print("received CMD ECHO packet: ", message)
            tokens = message.split(':')
            replyindex = message.find(":#^#")
            good = (len(tokens) >= 3) and (replyindex > 0)
            if (good):
                cmdstr = message[(replyindex + 4):]
                recvack = (tokens[0] == "RECV")
                cmdack = (tokens[0] == "CMD")
                goodstatus = (tokens[1] == "OK")
                badstatus = (tokens[1] == "FAIL")
                try:
                    idcode = int(tokens[2])
                except:
                    idcode = -9999
                    good = False

                good = good and (goodstatus or badstatus) and (recvack or cmdack)
                if (good):

                    try:
                        idx = self.pending_recv_ack.index(idcode)
                        self.pending_recv_ack.pop(idx)
                        good = True
                    except:
                        good = cmdack

                    if (cmdack):
                        try:
                            idx = self.pending_cmd_ack.index(idcode)
                            self.pending_cmd_ack.pop(idx)
                            # good = good and True
                        except:
                            good = False

                    if (cmdack):
                        f = self.customcallback.get(idcode)
                        if (f is not None):
                            f(goodstatus, cmdstr)
                            self.customcallback.pop(idcode)

                    for f in self.callbacks:
                        f(idcode, cmdstr, recvack, cmdack, goodstatus)

        self.lock.release()
