import os.path
import yaml
from PyQt5 import QtCore
import logging
import gc
import datetime

from .tab_commands.show_statemachine import SHOWGSE_CommandState, SHOWHK_Miscellaneous, SHOWHK_Temperatures, SHOWHK_Attitude
from .tab_commands.showgse_command_gui import SHOWGSE_CommandGui
from .tab_temperatures.showgse_temperatures_gui import SHOWGSE_TemperatureGui
from .tab_filelistings.showgse_filesgui import SHOWGSE_FileListingGui
from .tab_attitude.showgse_attitude_gui import SHOWGSE_AttitudeGui
from .htmldecode.showgse_networkio import SHOWSGE_NetworkIO
from . import mainapplication


# ------------------------------------------------------------------------------
#           read_science_modes
# ------------------------------------------------------------------------------

def read_science_modes():
    configuration = None
    fullname = __file__  # Get the filename of this module
    [basedir, spare] = os.path.split(fullname)  # split out the list_of_files
    fullname = basedir + os.sep + 'scienceprograms' + os.sep + 'sciencemodes.yaml'
    try:
        f = open(fullname, 'rt')
        configuration = yaml.load(f)
        f.close()
    except Exception as inst:
        print(type(inst))  # the exception instance
        print(inst.args)  # arguments stored in .args
        print(inst)  # __str__ allows args to be printed directly,
        raise RuntimeError(
            'Error reading the YAML science programs configuration file for the SHOW GSE from file %s' % (fullname,))
    return configuration

    # ------------------------------------------------------------------------------
    #               def load_config(self):
    # ------------------------------------------------------------------------------


def read_gse_config():
    configuration = None
    fullname = __file__  # Get the filename of this module
    [basedir, spare] = os.path.split(fullname)  # split out the list_of_files
    fullname = basedir + os.sep + 'scienceprograms' + os.sep + 'gse_config.yaml'
    try:
        f = open(fullname, 'rt')
        configuration = yaml.load(f)
        f.close()
    except Exception as inst:
        print(type(inst))  # the exception instance
        print(inst.args)  # arguments stored in .args
        print(inst)  # __str__ allows args to be printed directly,
        raise RuntimeError(
            'Error reading the YAML gse configuration file for the SHOW GSE from file %s' % (fullname,))
    return configuration


# ------------------------------------------------------------------------------
#           GSE_MainGlobalObject
# ------------------------------------------------------------------------------

class GSE_MainGlobalObject:

    def __init__(self, qtapp: mainapplication.GSE_MainApplication):
        self.gui_app = qtapp
        self.gse_config = read_gse_config()

        utc = datetime.datetime.utcnow()
        datestr = utc.strftime('%Y-%m-%d')
        dirname = self.gse_config['DEFAULTS']["base_output_dir"] + os.sep + datestr + os.sep + 'telemetry'
        if (not os.path.exists(dirname)):
            os.makedirs(dirname)
        loggername = dirname + os.sep + 'show_log_' + datestr + '_00-00.txt'
        logging.basicConfig(filename=loggername, filemode='a', level=logging.INFO, format='%(asctime)s  %(levelname)s: %(message)s ')
        logging.info('SHOW GSE start at %s', str(utc))

        self.network_io = SHOWSGE_NetworkIO(self.gui_app, self.gse_config)

        self.sciencemodes = read_science_modes()
        self.hktemps = SHOWHK_Temperatures(self.network_io, self.gse_config)
        self.miscellaneous = SHOWHK_Miscellaneous(self.network_io)
        self.commandstate = SHOWGSE_CommandState(self, self.network_io)
        self.hkattitude = SHOWHK_Attitude(self.network_io, self.gse_config)

        self.temperature_gui = SHOWGSE_TemperatureGui(self)
        self.command_gui = SHOWGSE_CommandGui(self, self.temperature_gui)
        self.filelisting_gui = SHOWGSE_FileListingGui(self)
        self.attitude_gui = SHOWGSE_AttitudeGui(self, self.gse_config)

        self.start_loop_timer()
        self.start_html_refresh_timer()
        self.start_attitude_refresh_timer()
        self.start_show_telemetry_html_refresh_timer()
        self.network_io.networkthreads['show_html'].start_html_thread()

    # ------------------------------------------------------------------------------
    #           SHOWSGE_TemperatureMonitorThread::start_loop_timer
    # ------------------------------------------------------------------------------

    def start_loop_timer(self):
        """ Start the timer used to update displays and settings from all the
        values acquired in the last few seconds. This guarantees that things keep synced up
        """

        loop_refresh_rate = self.gse_config['timers']['graph_refresh_rate']
        self.loop_timer = QtCore.QTimer(self.gui_app)
        self.loop_timer.setSingleShot(False)
        interval = 5000 if (loop_refresh_rate == 0.0) else (1000 * loop_refresh_rate)
        self.loop_timer.setInterval(int(interval))
        self.loop_timer.setTimerType(QtCore.Qt.CoarseTimer)
        self.loop_timer.timeout.connect(self.timer_loop)
        self.loop_timer.start(int(interval))

    # ------------------------------------------------------------------------------
    #           SHOWSGE_TemperatureMonitorThread::start_html_refresh_timer
    # ------------------------------------------------------------------------------

    def start_html_refresh_timer(self):
        """
        We dont really need the HTML for temperatures but we should poll SHOW every few tens of seconds
        so it doesnt think its lost a connection. And if we are going to poll then we may as well take the
        temperatures from the returned HTML

        :return:
        """
        refresh_rate = self.gse_config['timers']['html_refresh_rate']
        self.html_timer = QtCore.QTimer(self.gui_app)
        self.html_timer.setSingleShot(False)
        interval = 40000 if (refresh_rate == 0.0) else (1000 * refresh_rate)
        self.html_timer.setInterval(int(interval))
        self.html_timer.setTimerType(QtCore.Qt.CoarseTimer)
        self.html_timer.timeout.connect(self.update_mainpage_html)
        self.html_timer.start(int(interval))  # start the time the calls start_html_thread

    # ------------------------------------------------------------------------------
    #           SHOWSGE_TemperatureMonitorThread::start_html_refresh_timer
    # ------------------------------------------------------------------------------

    def start_attitude_refresh_timer(self):
        """
        Loop timer for the attitude system. We poll the web page every 0.5 seconds looking for change

        :return:
        """
        refresh_rate = self.gse_config['timers']['attitude_html_refresh_rate']
        self.attitude_html_timer = QtCore.QTimer(self.gui_app)
        self.attitude_html_timer.setSingleShot(False)
        interval = 1000 * refresh_rate
        self.attitude_html_timer.setInterval(int(interval))
        self.attitude_html_timer.setTimerType(QtCore.Qt.CoarseTimer)
        self.attitude_html_timer.timeout.connect(self.update_attitude_html)
        self.attitude_html_timer.start(int(interval))  # start the time the calls start_html_thread

    # ------------------------------------------------------------------------------
    #           SHOWSGE_TemperatureMonitorThread::start_html_refresh_timer
    # ------------------------------------------------------------------------------

    def start_show_telemetry_html_refresh_timer(self):
        """
        We dont really need the HTML for temperatures but we should poll SHOW every few tens of seconds
        so it doesnt think its lost a connection. And if we are going to poll then we may as well take the
        temperatures from the returned HTML

        :return:
        """
        #       refresh_rate = self.gse_config['timers']['html_refresh_rate']
        self.html_show_telem_timer = QtCore.QTimer(self.gui_app)
        self.html_show_telem_timer.setSingleShot(False)
        interval = 1000
        self.html_show_telem_timer.setInterval(int(interval))
        self.html_show_telem_timer.setTimerType(QtCore.Qt.CoarseTimer)
        self.html_show_telem_timer.timeout.connect(self.update_showtelemetry_html)
        self.html_show_telem_timer.start(int(interval))  # start the time the calls start_html_thread

    # ------------------------------------------------------------------------------
    #           SHOWSGE_TemperatureMonitorThread::start_html_thread
    # ------------------------------------------------------------------------------

    @QtCore.pyqtSlot()
    def update_mainpage_html(self):
        """
        The HTML polling timer calls this function at the end of each HTML timer loop.
        This function initiates the fetch of the ShowMainPage.py from SHOW.
        """
        self.network_io.networkthreads['show_html'].start_html_thread()

    # ------------------------------------------------------------------------------
    #           SHOWSGE_TemperatureMonitorThread::start_html_thread
    # ------------------------------------------------------------------------------

    @QtCore.pyqtSlot()
    def update_attitude_html(self):
        """
        The HTML polling timer calls this function at the end of each HTML timer loop.
        This function initiates the fetch of the ShowMainPage.py from SHOW.
        """
        self.network_io.networkthreads['navdat_url'].start_html_thread()

    # ------------------------------------------------------------------------------
    #           SHOWSGE_TemperatureMonitorThread::start_html_thread
    # ------------------------------------------------------------------------------

    @QtCore.pyqtSlot()
    def update_showtelemetry_html(self):
        """
        The HTML polling timer calls this function at the end of each HTML timer loop.
        This function initiates the fetch of the ShowMainPage.py from SHOW.
        """
        self.network_io.networkthreads['nasdat_url'].start_html_thread()

    # ------------------------------------------------------------------------------
    #           SHOWSGE_TemperatureMonitorThread::timer_loop
    # ------------------------------------------------------------------------------
    #

    @QtCore.pyqtSlot()
    def timer_loop(self):
        """ This the time loop that fires every few seconds. Its main purpose
        is to ensure housekeeping displays are updated and check that udp threads have not crashed.
        are restarted if they have crashed
        """

        self.temperature_gui.update_plots()
        self.attitude_gui.update_plots()
        self.network_io.check_threads()
        gc.collect()
