import pysftp
import numpy as np
from PyQt5 import QtWidgets
from .yaml_configuration import Yaml_Configuration


class argol_ftp(object):
    """
    This is a helper class which handles SFTP connections to the flight computer. Specifically, it aids in the ftp
    communication through the medium of a QTreeWidget. This class is intended to be
    instantiated in threads/functions which have a limited lifetime and call the destructor when the immediate task
    is done. A more permanent life time of this class is possible but be aware of the constant connection in terms of
    both bandwidth usage as well as avoiding multiple/redundant connections.
    """
    def __init__(self, ip_addr: str) -> None:

        self.ftp_network_settings: Yaml_Configuration = Yaml_Configuration('network.yaml').configuration['FTP']
        self.science_settings: Yaml_Configuration = Yaml_Configuration('science_image.yaml').configuration['SETTINGS']

        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None  # TODO: This currently disables host key checking, try fixing that.
        try:
            self.ftp_connection = pysftp.Connection(host=ip_addr,
                                                    username=self.ftp_network_settings["username"],
                                                    password=self.ftp_network_settings["password"],
                                                    port=self.ftp_network_settings["port"], cnopts=cnopts)
        except Exception as err:
            self.ftp_connection = str(err)  # Assign the error to the connection, can check if good connection with isinstance

    def refresh_remote_dir_in_tree(self, tree: QtWidgets.QTreeWidget) -> None:
        # This function takes in a QtreeWidget and will populate it with all directories in 'ftp_image_dir'
        dir_list = self.__pull_all_sub_dir__(self.science_settings['ftp_image_dir'])
        self.__populate_QTree_with_remote_dir__(tree, dir_list)

    def populate_dir_with_images(self, tree: QtWidgets) -> None:
        # This function takes a tree widget and will populate the selected branch of the tree with with all files
        # inside of it that are not also directories.
        image_list = []
        if isinstance(self.ftp_connection, pysftp.Connection):
            selected_path = self.science_settings['ftp_image_dir'] + r'/' + self.find_path_of_selected_item(tree)
            inside_dir = self.ftp_connection.listdir(selected_path)
            for item in inside_dir:
                if not self.ftp_connection.isdir(selected_path + r'/' + item):
                    image_list.append(item)
        for file in image_list:
            QtWidgets.QTreeWidgetItem(tree.selectedItems()[0], [file])

    def get_newest_file_path(self, next_path: str, looking_for_strip: bool=False) -> str:
        # This function just gets the path to the newest thing within the initial starting dir (the initial input).
        # There is no guarantee that this is a image.
        #
        # If anything has "strip_" in its name in the current folder then this function will proceed to
        # exclusively look at or or ignore the "strip_" folders based on looking_for_strip. This is done for ALIV3
        # science mode streaming to stream only strip versions or full versions of images selectively.
        newest_time = 0
        newest_path = ''
        if isinstance(self.ftp_connection, pysftp.Connection):
            attrs = self.ftp_connection.listdir_attr(next_path)
            # Look through names and see if "strip_" is there. If it is begin the selection of either only
            # "strip_" items or no "strip_" items.
            strip_folder_here = np.any(["strip_" in x.filename for x in attrs])
            for attr in attrs:
                if not strip_folder_here:
                    # Just get newest thing, no strip condition
                    if attr.st_mtime > newest_time:
                        newest_time = attr.st_mtime
                        newest_path = next_path + r'/' + attr.filename
                else:
                    # Check strip condition
                    if (attr.st_mtime > newest_time) and (looking_for_strip == ("strip_" in attr.filename)):
                        newest_time = attr.st_mtime
                        newest_path = next_path + r'/' + attr.filename
            if self.ftp_connection.isdir(newest_path):
                newest_path = self.get_newest_file_path(newest_path, looking_for_strip)
            if len(attrs) == 0:
                # Found the end
                newest_path = next_path
        return newest_path

    def find_path_of_selected_item(self, tree: QtWidgets) -> str:
        # This function takes in a tree widget and returns the full path of the selected item.
        selected_full_path = ""
        current_item = tree.selectedItems()[0]
        while current_item.parent() is not None:
            selected_full_path = current_item.parent().text(0) + r'/' + selected_full_path
            current_item = current_item.parent()
        selected_full_path += tree.selectedItems()[0].text(0)
        return selected_full_path

    def __pull_all_sub_dir__(self, in_path: str) -> list:
        dir_list = []
        if isinstance(self.ftp_connection, pysftp.Connection):
            if self.ftp_connection.isdir(in_path):
                inside_dir = self.ftp_connection.listdir(in_path)
                for item in inside_dir:
                    next_path = in_path + r'/' + item
                    if self.ftp_connection.isdir(next_path):
                        dir_list.append(item)
                        next_list = self.__pull_all_sub_dir__(next_path)
                        if len(next_list) > 0:
                            dir_list.append(next_list)
        return dir_list

    def __populate_QTree_with_remote_dir__(self, tree: QtWidgets.QTreeWidget, dir_list: list) -> None:
        parent_itm = tree
        for idx in range(len(dir_list)):
            if not isinstance(dir_list[idx], list):
                parent_itm = QtWidgets.QTreeWidgetItem(tree, [dir_list[idx]])
            else:
                self.__populate_QTree_with_remote_dir__(parent_itm, dir_list[idx])

    def __del__(self) -> None:
        # This has been confirmed to get called when this class is used as intended
        if isinstance(self.ftp_connection, pysftp.Connection):
            self.ftp_connection.close()
