import logging
from typing import Dict, Any, List, Tuple
from PyQt5 import QtCore, QtWidgets
import os
import os.path
import datetime
from .widgets.sensor_variable_array import SensorVariableUI_Array
from .widgets.tab_widget_manager import Tab_Widget_Manager
from .network.udp_listener import UDP_Listener
from .yaml_configuration import Yaml_Configuration
from .cmd import CMD_Manager
from .widgets.sensor_variable_widgets import TrafficLight, SensorVariableUI_String
from .widgets.tab_widget_manager import Tab_Widget_Manager


# from .hk_logger import hk_logger

# -----------------------------------------------------------------------------------------------------------------------
#           HK_Listener
# -----------------------------------------------------------------------------------------------------------------------

class HK_Listener:
    '''
    Class listens for housekeeping packets broadcast over to a UDP socket and updates data stuctures and GUI elements.
    This class assumes the housekeeping packets are specified as a string of comma seperated values. This is compatible with
    the housekeeping stratgey used in ARG C++ firmware libraries.
    '''

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, qtapp: QtWidgets.QApplication, tab_widget_manager: Tab_Widget_Manager, cmd_mgr: CMD_Manager, sensorvariablelist: SensorVariableUI_Array, config_dir=None):

        config = Yaml_Configuration('network.yaml', basedir=config_dir)
        guilogconfig = Yaml_Configuration('gse_logger.yaml', basedir=config_dir)
        self.udp: UDP_Listener = UDP_Listener(qtapp, config.configuration['Network'], config.configuration['HK'])
        self.cmd_object: CMD_Manager = cmd_mgr
        self.tab_widget_manager: Tab_Widget_Manager = tab_widget_manager
        self.firmware_ip: QtWidgets.QLabel = getattr(qtapp.ui, 'FIRMWARE_IP_ADDRESS')
        self.hk_udp_port: SensorVariableUI_String = SensorVariableUI_String('HK_UDP_PORT', 'str', qtapp, tab_widget_manager, None, options={'tabicon': "voltages_currents:HK"})
        self.command_pending: int = 0
        self.hkvariables: SensorVariableUI_Array = sensorvariablelist
        self.hkvalues: Dict[str, str] = {}
        self.hkfields: List[Tuple[str, str]] = []
        self.do_autotimecheck: bool = True  # check that firmware time is reasonably close to the GSE time
        self.do_gse_addr_notify: bool = config.configuration['HK']['gse_addr_notify']
        #       self.hk_logger          : hk_logger               = hk_logger()
        self.log_basedirectory: str = guilogconfig.configuration['gse_logger_path']
        self.log_basedirectory = os.path.normpath(self.log_basedirectory)

        self.udp.add_listener(self.hk_callback)
        self.hk_udp_port.set_value(str(self.udp.udp_port_inst))

    # ------------------------------------------------------------------------------
    #           firmware_ip_address
    # ------------------------------------------------------------------------------

    def firmware_ip_address(self) -> str:
        '''
        Returns the network address of the firmware. It is generated when a housekeeping UDP packet is receivedthe
        '''
        return self.udp.last_ip_address_recvd

    # ------------------------------------------------------------------------------
    #           record_message
    # ------------------------------------------------------------------------------

    def record_message(self, message: str, file_prefix: str = "gsehk_"):
        '''
        Writes the current housekeeping string to a file on the GSe.
        '''

        tnow = datetime.datetime.utcnow()
        tmes = tnow.strftime("%Y%m%d")
        basefolder = self.log_basedirectory + os.sep + tmes
        basefolder = os.path.normpath(basefolder)
        os.makedirs(basefolder, exist_ok=True)
        filename = basefolder + os.sep + file_prefix + tmes + ".txt"
        f = open(filename, 'a+t')
        f.write("%s\n" % (message,))
        f.close()

    # ------------------------------------------------------------------------------
    #           SendGSENetworkAddress
    # ------------------------------------------------------------------------------

    def SendGSENetworkAddress(self):
        '''
        A function used to send the network address of the GSE to the firmware. This allows the firmware to send UDP
        packets to a specific UDP socket rather than broadcast UDP packets which may get blocked if they have to have
        pass through a network-to-network gateway. This assumes the firmware is running the ARG C++ firmware library.
        '''

        if self.do_gse_addr_notify:
            command = "SYSCTL SETGSEADDR " + self.cmd_object.cmd_send.local_address()
            print('Issuing command [{}]'.format(command))
            self.cmd_object.send_command(command)

    # ------------------------------------------------------------------------------
    #           hk_callback
    # ------------------------------------------------------------------------------

    def hk_callback(self, udp_packet: str):
        '''
        The callback function used to process housekeeping packets received from the C++ firmware.
        '''

        self.tab_widget_manager.reset_status("HK")  # Reset all the HK statuses for the icons in the tabs on the top row
        if (udp_packet is None) or (len(udp_packet) == 0):
            self.hk_udp_port.status = TrafficLight.Red
            self.SendGSENetworkAddress()
        else:
            if (self.firmware_ip is not None):
                self.firmware_ip.setText(self.firmware_ip_address())

            self.hk_udp_port.status = TrafficLight.Green
            tabfield = 'HK'
            self.tab_widget_manager.reset_status(tabfield)  # reset the status of the icons in the corner of the tabbed dialogs

            if (len(self.hkfields) == 0):  # if the housekeeping fields are undefined
                if (self.command_pending == 0):  # Then if we have not already initiated a query
                    self.command_pending = 1  # Initiate a query from the firmware
                    self.cmd_object.send_command('HK PRINTHEADER', self.hk_printheaders_callback)  # to get the Housekeeping header from the firmware
                else:
                    self.command_pending += 1  # Wait for 6-8 seconds/ HK packets before trying the command again
                    if (self.command_pending > 6):
                        self.command_pending = 0  # If query in progress then timout in 6-8 seconds so we can try again
            else:
                self.record_message(udp_packet)
                tokens = udp_packet.split(',')
                good = len(tokens) == len(self.hkfields)
                if (not good):
                    logging.warning('hk.hk_callback, The number of tokens in the HK string does not math the number of fields in HK PRINTHEADER')
                    self.hkfields.clear()
                else:
                    for i in range(len(self.hkfields)):  # First we want to extract all the hk fields into a temporary dictionary
                        fielder = self.hkfields[i][0]  # so get the name each HK field sent (earlier) from the firmware
                        value = tokens[i]  # get the (string) value of this field which we have just received
                        self.hkvalues[fielder] = value  # and set the value into a dictionary

                    if (self.do_autotimecheck):  # only do this check once
                        utcstr = self.hkvalues.get('UTC')
                        if (utcstr is not None) and (len(utcstr) > 16):
                            # if (len(utcstr) == 23):     # ifTS sends time to milliseconds only
                            #    utcstr = utcstr + "000"
                            # if utcstr[10] == '-':       #iFTS include a '-' sign between the days and hours.
                            #    utcstr = utcstr[0:10] + ' ' + utcstr[11:]
                            firmware_utc = datetime.datetime.strptime(utcstr, "%Y-%m-%d %H:%M:%S.%f")
                            gse_utc = datetime.datetime.utcnow()
                            dt = abs(gse_utc.timestamp() - firmware_utc.timestamp())
                            if (dt > 10):
                                print("GSE and Firwmare are not synchronized in UTC. They differ by ", dt, " seconds. Sending command to change firmware time")
                                command = "SYSCTL SETSYSTEMTIME " + gse_utc.strftime("%Y-%m-%d %H:%M:%S.%f")
                                self.cmd_object.send_command(command)

                    for aname, entry in self.hkvariables.items():  # Now loop over all of the sensor variables
                        name = entry.hk_name  # get the housekeeping name of this sensor variable
                        value = self.hkvalues.get(entry.hk_name)  # and look it up in our local dictionary
                        if (value is not None):  # if we found a value
                            entry.set_value(value)  # then set the value
                            self.tab_widget_manager.update_tab_status(entry.tabiconname, tabfield, entry.status)

                # Do logging
                # for i in range(len(self.hkfields)):
                #    field = self.hkfields[i][0]
                #    value = self.hkvalues[field]
                #    self.hk_logger.log_value(field, value, self.hkvalues.get('UTC'))

            # print("Received UDP packet: ", udp_packet)

    # ------------------------------------------------------------------------------
    #           hk_printheaders_callback
    # ------------------------------------------------------------------------------

    def hk_printheaders_callback(self, success: bool, header_fields: str):
        '''
        Specialized function used to record the headers for each of the comma delimited fields in a house keeping
        record. This supports the ARG C++ firmware.
        '''

        #        print("into hk_printheaders_callback")                             # we have received an response to our special command
        self.command_pending = 0  # so reset the command pending flag
        if not success:  # and if it failed
            self.hkfields.clear()  # then clear our decding fields so we can try again
        else:  # otherwise
            tnow = datetime.datetime.utcnow()
            tmes = tnow.strftime("%Y%m%d-%H%M%S, ")

            self.record_message(tmes + header_fields, file_prefix="gsefld_")
            tokens = header_fields.split(',')  # get the number of tokens sent down in the header
            for token in tokens:  # and insert
                # (name,type) = token.split(':')                              # each token (name and type)
                name = token.strip()
                self.hkfields.append((name, "variable"))  # into our description of the fields
