from typing import Any, Sequence, Dict, List, Union, NamedTuple, Tuple, TypeVar
import math
from lxml import html
import logging
import datetime


# from .show_statemachine import SHOWGSE_SystemState, SHOWGSE_CommandStateVariableBase, SHOWGSE_CommandState


# -----------------------------------------------------------------------------------------
#               readHTMLTree
# -----------------------------------------------------------------------------------------

def readHTMLTree(filename: str) -> html.HtmlElement:
    f = open(filename, 'rt')
    content = f.read()
    f.close()
    tree = html.fromstring(content)
    return tree


# ------------------------------------------------------------------------------
#           findentry
# ------------------------------------------------------------------------------

def findentry(tdlist: List[html.HtmlElement], name: str) -> int:
    n = len(tdlist)
    value = None
    for i in range(n):
        s = tdlist[i].text_content()
        if s.find(name) == 0:
            value = tdlist[i + 1].text_content()
            break
    if (value is None):
        print('decodehtml.findentry, could not find entry for <', name, '>')
    return value


# ------------------------------------------------------------------------------
#           find_multiline_entry
# ------------------------------------------------------------------------------

def find_multiline_entry(tdlist: List[html.HtmlElement], name: str) -> int:
    n = len(tdlist)
    value = None
    for i in range(n):
        s = tdlist[i].text_content()
        if s.find(name) == 0:
            thistd = tdlist[i + 1]
            value = [thistd.text]
            children = thistd.getchildren()
            for child in children:
                l = child.tail
                if len(l) > 0:
                    value.append(l)
            break
    if (value is None):
        print('decodehtml.find_multiline_entry, could not find entry for <', name, '>')
    return value


# ------------------------------------------------------------------------------
#           findutcentry
# ------------------------------------------------------------------------------

def findutcentry(td: List[html.HtmlElement]) -> datetime:
    utc = findentry(td, 'System Time (UTC)')
    return datetime.datetime.strptime(utc, "%Y-%m-%d %H:%M:%S")


# ------------------------------------------------------------------------------
#           findfloat
# ------------------------------------------------------------------------------

def findfloat(td: List[html.HtmlElement], name: str) -> float:
    value = findentry(td, name)
    if (value is not None) and (value.find('N/A') < 0):
        s = value.split()
        f = float(s[0])
    else:
        f = math.nan
    return f


# ------------------------------------------------------------------------------
#           findstring
# ------------------------------------------------------------------------------

def findstring(td: List[html.HtmlElement], name: str) -> float:
    value = findentry(td, name)
    if (value is None):
        value = 'NOT FOUND'
    #    print(name, ' = ', value)
    return value


# ------------------------------------------------------------------------------
#           find_select_string
# ------------------------------------------------------------------------------

def find_select_string(tselect: List[html.HtmlElement], name: str) -> str:
    value = None
    for selectentry in tselect:  # Scan through the list of HTML <select> .... </select> elelemnts
        s = selectentry.attrib.get('name')  # Does this select enetry have the name we are lookinmg for?
        if (s is not None) and (s == name):  # If it does
            options = selectentry.findall(r'.//option')  # Then find all the <option> ... </option> entries  in this select segment
            for optionentry in options:  # for each option entry
                selstr = optionentry.attrib.get('selected')  # see if it hass the selected attribute value
                if (selstr is not None) and (selstr == 'selected'):  # if it does then
                    value = str(optionentry.text_content())  # Get the text content displayed
                    break  # and we are done looping options
            break  # and we are done looping select elements
    if value is None:  # did we find the required value (we should)
        print("No select string value found on management page for ", name)
        value = "NOT FOUND"
    return value


# ------------------------------------------------------------------------------
#           find_input_string
# ------------------------------------------------------------------------------

def find_input_string(tinput: List[html.HtmlElement], name: str) -> str:
    value = None
    for inputentry in tinput:  # Scan through the list of HTML <select> .... </select> elelemnts
        s = inputentry.attrib.get('name')  # Does this select enetry have the name we are lookinmg for?
        if (s is not None) and (s == name):  # If it does
            value = inputentry.attrib.get('value')
            break
    if value is None:  # did we find the required value (we should)
        print("No select string value found on management page for ", name)
        value = "NOT FOUND"
    return value


# ------------------------------------------------------------------------------
#           find_input_float
# ------------------------------------------------------------------------------

def find_input_float(tinput: List[html.HtmlElement], name: str) -> float:
    value = None
    for inputentry in tinput:  # Scan through the list of HTML <select> .... </select> elelemnts
        s = inputentry.attrib.get('name')  # Does this select enetry have the name we are lookinmg for?
        if (s is not None) and (s == name):  # If it does
            try:
                value = float(str(inputentry.attrib.get("value")))
            except:
                print('error decoding float from string', inputentry.attrib.get("value"))
                value = math.nan
            break
    if value is None:  # did we find the required value (we should)
        print("No select string value found on management page for ", name)
        value = math.nan

    return value


# ------------------------------------------------------------------------------
#           decode_mainpage
# ------------------------------------------------------------------------------

def decode_mainpage(content: str) -> Tuple[
    Dict[str, Union[str, float, datetime.datetime]], Dict[str, Union[str, float, datetime.datetime]], Dict[
        str, Union[str, float, datetime.datetime]]]:
    """
    Decodes the information from the mainpage.py generated by the SHOW instrument.
    The fields are decoded into 1) HK temperatures, 2) Miscellaneous and 3) Command Status
    fields. These sections can then be given to the current instrument state and checked against the
    expected values.
    
    :param tree: The HTML object generated from parsing the HTML retuned from SHOW mainpage.py  
    :return: A tuple of three dictionaries [ HK Temps, Misccellaneous, Command Status]
    """

    if (content is not None) and (len(content) > 0):
        tree = html.fromstring(content)
        td = tree.findall(r'.//td')
    else:
        td = []
        print('HTML request failed to receive show mainpage from SHOW')
        logging.warning('HTML request failed to receive show mainpage from SHOW')

    if len(td) > 0:
        hktemps = {}
        hktemps["utc"] = findutcentry(td)
        hktemps["shs_top"] = findfloat(td, 'Interferometer Top Temperature')
        hktemps["shs_bot"] = findfloat(td, 'Interferometer Bottom Temperature')
        hktemps["optobox"] = findfloat(td, 'Opto-Box Inner Enclosure Temperature')
        hktemps["case"] = findfloat(td, 'Camera Case Temperature')
        hktemps["ebox"] = findfloat(td, 'E-Box Temperature')
        hktemps["wingpod"] = findfloat(td, 'WingPod Window Temperature Sensor')
        hktemps["periscope"] = findfloat(td, 'Periscope Temperature Sensor')
        hktemps["fpa"] = findfloat(td, 'Camera Sensor FPA Temperature')
        hktemps["pcb"] = findfloat(td, 'Camera Sensor Circuit Temperature')
        hktemps["q7"] = findfloat(td, 'Q7 Temperature')
        hktemps["heater_top_power"] = findfloat(td, 'Heater Top Power')
        hktemps["heater_bot_power"] = findfloat(td, 'Heater Bottom Power')

        miscellaneous = {}
        miscellaneous["utc"] = hktemps["utc"]
        miscellaneous["reported_gain"] = findstring(td, 'Reported Camera Gain Mode')
        miscellaneous["reported_exposuretime"] = findfloat(td, 'Reported Camera Exposure Time')
        miscellaneous["storage_space"] = findstring(td, 'Storage Space (Free / Total)')
        miscellaneous["data_transferred"] = findstring(td, 'Amount of data transferred (since power-on)')
        miscellaneous["warnings"] = find_multiline_entry(td, 'Warning(s)')
        miscellaneous["lowlevefaults"] = find_multiline_entry(td, 'Low level system fault(s)')

        commandstate = {}
        commandstate["science_mode"] = findstring(td, 'Images description comment')
        commandstate["gain_selection"] = findstring(td, 'Gain Selection (High / Low)')
        commandstate["nuc_mode"] = findstring(td, 'NUC Mode')
        commandstate["agc_selection"] = findstring(td, 'AGC Selection (ON / OFF)')
        commandstate["agc_region"] = findstring(td, 'AGC Region (offsetY|sizeY|offsetX|sizeX)')
        commandstate["readout_mode"] = findstring(td, 'Readout Mode Selection (ITR / IWR)')
        commandstate["tec_control"] = findstring(td, 'Camera TEC Control Selection (ON / OFF)')
        commandstate["tec_setpoint"] = findfloat(td, 'Camera TEC Temperature')
        commandstate["exposure_mode"] = findstring(td, 'Exposure Mode (Auto / Manual / Vector)')
        #    commandstate["exposure_vector"]    =
        commandstate["exposure_time"] = findstring(td, 'Exposure Time')
        commandstate["image_acq_at_pwr"] = findstring(td, 'Image Acquisition At PowerOn (On/Off)')
        commandstate["image_acq_rate"] = findfloat(td, 'Image Acquisition Rate (0 - 4.0 Hz)')
        commandstate["tm_refresh_rate"] = findfloat(td, 'Telemetry Refresh Rate (0 - 1.0 Hz)')
        commandstate["timeout_secs"] = findfloat(td, 'Ground Communication Timeout (60 - 600 Seconds)')
        commandstate["web_refresh_rate"] = findfloat(td, 'Web Interface Auto-Refresh Rate (0 - 200 Seconds)')
        commandstate["show_image_sample"] = findstring(td, 'Show sample image on web page (On/Off)')
        commandstate["faillight_override"] = findstring(td, 'Fail Light Override (Auto / Force ON /Force OFF)')

        for l in miscellaneous["warnings"]:
            if l is not None:
                logging.info(' SHOW WARNING MESSAGE:%s', l)

        for l in miscellaneous["lowlevefaults"]:
            if l is not None:
                logging.info(' SHOW LOW-LEVEL FAULT: %s', l)

    else:
        hktemps = {}
        miscellaneous = {}
        commandstate = {}

    return [hktemps, miscellaneous, commandstate]


# ------------------------------------------------------------------------------
#           decode_management_page
# ------------------------------------------------------------------------------

def decode_management_page(content: str) -> Dict[str, Union[str, float, datetime.datetime]]:
    """
    Decodes the information from the management page generated by the SHOW instrument.
    The fields are decoded into (1) Command Status fields. This sections can then be 
    given to the current instrument state and checked against the expected values.
    
    :param tree: The HTML object generated from parsing the HTML retuned from SHOW management page
    :return: The dictionary of values that can be applied to the SHOW GSE Command State
    """

    commandstate = {}
    if (content is not None) and (len(content) > 0):
        tree = html.fromstring(content)
        tselect = tree.findall(r'.//select')
        tinput = tree.findall(r'.//input')
        commandstate["science_mode"] = find_input_string(tinput, 'ImagesDescription')
        commandstate["gain_selection"] = find_select_string(tselect, 'GainSelection')
        commandstate["nuc_mode"] = find_select_string(tselect, 'NucMode')
        #    commandstate["agc_selection"]      = find_select_string (tselect, 'AgcSelection')
        #    commandstate["agc_region"]         = find_input_string  (tinput,  'AgcRegion')
        #    commandstate["readout_mode"]       = find_select_string (tselect, 'ReadoutMode')
        commandstate["tec_control"] = find_select_string(tselect, 'CameraTecControl')
        commandstate["tec_setpoint"] = find_input_float(tinput, 'CameraTecTemperature')
        commandstate["exposure_mode"] = find_select_string(tselect, 'ExposureMode')
        commandstate["exposure_vector"] = find_input_string(tinput, 'ExposureVector')
        commandstate["exposure_time"] = find_input_float(tinput, 'ExposureTime')
        commandstate["image_acq_at_pwr"] = find_select_string(tselect, 'ImageAcquisitionPowerON')
        commandstate["image_acq_rate"] = find_input_float(tinput, 'ImageAcquisitionRate')
        commandstate["tm_refresh_rate"] = find_input_float(tinput, 'TelemetryRate')
        commandstate["timeout_secs"] = find_input_float(tinput, 'CommunicationTimeoutSeconds')
        commandstate["web_refresh_rate"] = find_input_float(tinput, 'WebRefreshSeconds')
        commandstate["show_image_sample"] = find_select_string(tselect, 'SampleImageInInterface')
        commandstate["faillight_override"] = find_select_string(tselect, 'FailOverride')
    else:
        print('HTML POST failed to receive a proper management page from SHOW')
        logging.warning('HTML POST failed to receive a proper management page from SHOW')
    return commandstate
