from typing import Dict, List, Any, Callable
from PyQt5 import QtCore, QtWidgets
import threading
import logging
import datetime
import socket
from .decodehtml import decode_mainpage
from .threadedhtml import ThreadedHTMLRequest


# ------------------------------------------------------------------------------
#           class HTML_Listener
# ------------------------------------------------------------------------------

class HTML_SHOW_MainPage_Listener(QtCore.QObject):
    """
    A class that listens to the HTML returned from the SHOW Main page and passes the information onto the listeners
    """

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, app: QtWidgets.QApplication, config: Dict[str, Any]):

        super().__init__()
        self.app = app  # type: QtWidgets.QApplication
        self.listeners = []  # type: List[Callable]
        self.name = "SHOW HTML"
        self.timeout = config['timeout']
        self.url = 'http://' + config["show_ip"] + '/cgi-bin/ShowMainPage.py'
        self.isdirty = False
        self.lock = threading.Lock()
        self.htmlthread = ThreadedHTMLRequest(app, config, maxtries=-1)
        self.htmlthread.dummydata = 'ABB SHOW CCTS.htm'

        self.htmlthread_isactive = False

    # ------------------------------------------------------------------------------
    #           add_listener
    # ------------------------------------------------------------------------------

    def add_listener(self, listenfunc: Callable):

        self.listeners.append(listenfunc)

    # ------------------------------------------------------------------------------
    #           check_network_thread
    # ------------------------------------------------------------------------------

    def check_network_thread(self):
        pass

    # ------------------------------------------------------------------------------
    #           SHOWSGE_TemperatureMonitorThread::process_html_request
    # ------------------------------------------------------------------------------

    def process_html_request(self, htmlcontent):

        if htmlcontent is not None:
            hktemps, miscellaneous, commandstate = decode_mainpage(htmlcontent)
            if (self.htmlthread.urlerror):  # If we had a URL error reading the HTML
                g = hktemps.get("utc")  # Then see if we have any entries. If we do then we read in testdata
                if (g is not None):  # so if we have data
                    hktemps["utc"] = datetime.datetime.utcnow()  # then adjust the UTC to now
                    miscellaneous["utc"] = datetime.datetime.utcnow()  # Do the same with the miscellaneous information

            for listenfunc in self.listeners:
                listenfunc(hktemps, miscellaneous, commandstate)

        self.lock.acquire()
        self.htmlthread_isactive = False
        self.lock.release()

    # ------------------------------------------------------------------------------
    #           start_html_thread
    # ------------------------------------------------------------------------------

    def start_html_thread(self):
        """
        The HTML polling timer calls this function at the end of each HTML timer loop.
        This function initiates the fetch of the ShowMainPage.py from SHOW.

        :return:
        """
        self.lock.acquire()
        if not self.htmlthread_isactive:
            self.htmlthread_isactive = True

            self.htmlthread.start_htmlfetch(self.url, self.process_html_request, timeout=self.timeout)
        self.lock.release()


# ------------------------------------------------------------------------------
#           class HTML_Listener
# ------------------------------------------------------------------------------

class HTML_Listener(QtCore.QObject):

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, app: QtWidgets.QApplication, config: Dict[str, Any]):

        super().__init__()
        self.app = app  # type: QtWidgets.QApplication
        self.config = config  # type: Dict[str,Any]
        self.name = config["name"]
        self.timeout = config['timeout']
        self.url = config["url"]
        self.url_on = config["url_on"]
        self.htmlthread = ThreadedHTMLRequest(app, config)
        self.listeners = []  # type: List[Callable]
        self.isdirty = False
        self.lock = threading.Lock()
        self.htmlthread_isactive = False

    # ------------------------------------------------------------------------------
    #           add_listener
    # ------------------------------------------------------------------------------

    def add_listener(self, listenfunc: Callable):

        self.listeners.append(listenfunc)

    # ------------------------------------------------------------------------------
    #           check_network_thread
    # ------------------------------------------------------------------------------

    def check_network_thread(self):
        pass

    # ------------------------------------------------------------------------------
    #           SHOWSGE_TemperatureMonitorThread::process_html_request
    # ------------------------------------------------------------------------------

    def process_html_request(self, htmlcontent):

        if (htmlcontent is not None):
            for listenfunc in self.listeners:
                listenfunc(htmlcontent)

        self.lock.acquire()
        self.htmlthread_isactive = False
        self.lock.release()

    # ------------------------------------------------------------------------------
    #           start_html_thread
    # ------------------------------------------------------------------------------

    def start_html_thread(self):
        """
        The HTML polling timer calls this function at the end of each HTML timer loop.
        This function initiates the fetch of the ShowMainPage.py from SHOW. Dont start a new request if one is already active

        :return:
        """
        if self.url_on:
            self.lock.acquire()
            if not self.htmlthread_isactive:
                self.htmlthread_isactive = True

                self.htmlthread.start_htmlfetch(self.url, self.process_html_request, timeout=self.timeout)
            self.lock.release()
