from typing import Dict, List, Any
from PyQt5 import QtCore, QtGui, QtWidgets
import logging
import urllib
import urllib.parse
import urllib.request
import threading
import os.path


# ------------------------------------------------------------------------------
#           ThreadedHTMLRequest
# ------------------------------------------------------------------------------

class ThreadedHTMLRequest(QtCore.QObject):

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, app: QtWidgets.QApplication, config: Dict[str, str], isbinaryfile=False, maxtries=10):

        super().__init__()
        self.isbinaryfile = isbinaryfile
        self.maxtries = maxtries
        self.numbad = 0
        self.app = app
        self.eventid = QtCore.QEvent.registerEventType()
        self.busy = False
        self.postvalues = None
        self.htmldata = None
        self.thread = None
        self.url = None
        self.urlerror = False
        self.use_dummy_data = config.get("use_dummydata")
        self.dummydata = None
        self.lock = threading.Lock()

    # ------------------------------------------------------------------------------
    #           start_htmlfetch
    # ------------------------------------------------------------------------------

    def start_htmlfetch(self, url: str, callbackmethod: Any, post_values: Dict[str, str] = None, timeout=10):
        if ((self.numbad < self.maxtries) or self.maxtries < 0):
            self.lock.acquire()
            self.busy = True
            self.postvalues = post_values
            self.timeout = timeout
            self.htmldata = None
            self.url = url
            self.callbackmethod = callbackmethod
            self.thread = threading.Thread(target=_localthread_requesthtml_and_wait_for_return, args=(self,))
            self.thread.start()

    # ------------------------------------------------------------------------------
    #           finish_htmlfetch
    # ------------------------------------------------------------------------------

    def finish_htmlfetch(self):
        self.lock.release()

        if self.urlerror:
            self.numbad += 1
            if (self.numbad >= self.maxtries) and (self.maxtries > 0):
                logging.warning('Max tries on HTML connection to %s reached. This connection is now disabled', self.url)
                print('Max tries on HTML connection to ', self.url, ' reached. This connection is now disabled')
        else:
            self.numbad = 0
        self.thread = 0  # delete the thread object
        self.callbackmethod(self.htmldata)

    # ------------------------------------------------------------------------------
    #           event
    # ------------------------------------------------------------------------------

    def event(self, thisevent: QtCore.QEvent) -> bool:
        if (thisevent.type() == self.eventid):
            thisevent.accept()
            self.finish_htmlfetch()
            return True
        return super().event(thisevent)


# ------------------------------------------------------------------------------
#           _localthread_post_HTML_and_wait_for_return
# ------------------------------------------------------------------------------

def _localthread_requesthtml_and_wait_for_return(threadrequest: ThreadedHTMLRequest):
    try:
        if threadrequest.postvalues is not None:
            data = urllib.parse.urlencode(threadrequest.postvalues)
            data = data.encode()  # data should be bytes
        else:
            data = None
        req = urllib.request.Request(threadrequest.url, data)
        h = urllib.request.urlopen(req, timeout=threadrequest.timeout)
        source = h.read()
        h.close()
        if (not threadrequest.isbinaryfile):
            threadrequest.htmldata = source.decode('utf-8')
        else:
            threadrequest.htmldata = source

        threadrequest.urlerror = False
    except (urllib.request.URLError, ConnectionResetError):
        threadrequest.htmldata = ''
        threadrequest.urlerror = True
        if (threadrequest.use_dummy_data) and (threadrequest.dummydata is not None) and (
                len(threadrequest.dummydata) > 0):
            fullname = __file__  # Get the filename of this module
            [basedir, spare] = os.path.split(fullname)  # split out the list_of_files
            [basedir, spare] = os.path.split(basedir)  # split out the list_of_files
            fullname = basedir + os.sep + 'testdata' + os.sep + threadrequest.dummydata
            f = open(fullname, 'rb')
            threadrequest.htmldata = f.read()
            f.close()

    threadrequest.busy = False
    event = QtCore.QEvent(threadrequest.eventid)
    threadrequest.app.postEvent(threadrequest, event)
