from typing import Dict, List, Any, Callable, Tuple
from PyQt5 import QtCore, QtWidgets
import threading
import logging
import socket
import netifaces
import re
import sys

if sys.platform == 'win32':
    import winreg


# ------------------------------------------------------------------------------
#           get_connection_name_from_guid
# ------------------------------------------------------------------------------

def get_connection_name_from_guid(iface_guids: List[str]) -> List[str]:
    """
    On windows converts a GUID network address into a user friendly string by reading the windows registry.

    Parameters
    ----------
    iface_guids: List[str]
        A list of GUIDS

    """
    assert(sys.platform == 'win32')
    iface_names = ['(unknown)' for i in range(len(iface_guids))]
    reg = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
    reg_key = winreg.OpenKey(reg, r'SYSTEM\CurrentControlSet\Control\Network\{4d36e972-e325-11ce-bfc1-08002be10318}')
    for i in range(len(iface_guids)):
        try:
            reg_subkey = winreg.OpenKey(reg_key, iface_guids[i] + r'\Connection')
            iface_names[i] = winreg.QueryValueEx(reg_subkey, 'Name')[0]
        except FileNotFoundError:
            pass
    return iface_names


# ------------------------------------------------------------------------------
#           mac_address_to_ip_and_broadcast
# ------------------------------------------------------------------------------
def mac_address_to_ip_and_broadcast(desiredmac: str) -> str:
    ipaddr = None
    m = re.match(r"\d+\.\d+\.\d+\.\d+", desiredmac)  # Do we have a regulr TCP/IP address eg 172.20.3.151
    if m is not None:
        ipaddr = desiredmac
    else:  # Not a TCP/IP. So do we have a MAC Address, e.g f8:b1:56:c3:64:f3
        m = re.match(r"\w+\:\w+\:\w+\:\w+\:\w+\:\w+", desiredmac)
        if m is not None:
            thismacaddr = desiredmac.upper()
            interfaces = netifaces.interfaces()
            for interface in interfaces:
                if ipaddr is None:
                    connection = netifaces.ifaddresses(interface)
                    mac = connection.get(netifaces.AF_LINK)
                    addr = connection.get(netifaces.AF_INET)
                    if (mac is not None) and (addr is not None):
                        if (mac[0]['addr'].upper() == thismacaddr):
                            ipaddr = addr[0]['addr']
                            break
        else:  # Not a MAC address so it is a Windows connection name
            interface_guids = netifaces.interfaces()
            connectionnames = get_connection_name_from_guid(interface_guids)
            targetname = desiredmac.lower()
            for idx, name in enumerate(connectionnames):
                if name.lower() == targetname:
                    guid = interface_guids[idx]
                    connection = netifaces.ifaddresses(guid)
                    mac = connection.get(netifaces.AF_LINK)
                    addr = connection.get(netifaces.AF_INET)
                    if (addr is not None):
                        ipaddr = addr[0]['addr']
                    break
    if (ipaddr is None):
        print("mac_address_to_ip_and_broadcast: Could not find a network connection on this machine with Address {}. Please check network is not disconnected.".format(desiredmac))

    return ipaddr


# ------------------------------------------------------------------------------
#           UDP_Listener
# ------------------------------------------------------------------------------

class UDP_Listener(QtCore.QObject):
    """
    A multi-threaded class that listens to messages coming in from UDP ports and calls any callable functions that
    have registered with the object through :meth:`add_listener`. This class operates in a PyQt5 environment and sends
    events to the main application to send UDP packet data from the UDP listening thread to the main GUI application.
    """

    # ------------------------------------------------------------------------------
    #          UDP_Listener:: __init__
    # ------------------------------------------------------------------------------

    def __init__(self, app: QtWidgets.QApplication,
                 networkconfig: Dict[str, Any],
                 config: Dict[str, Any]):

        super().__init__()

        ipaddress = mac_address_to_ip_and_broadcast(networkconfig['gse_interface'])

        self.app = app  # type: QtWidgets.QApplication
        self.udp_on = config['udp_on']
        self.name = config['name']
        self.udp_ipaddress = ipaddress
        self.udp_port_gse = config['udp_port_gse']
        self.udp_port_inst = config['udp_port_inst']
        self.udp_max_message = config['udp_max_message']
        self.timeout = config['timeout']

        self.listeners = []  # type: List[Callable]
        self.udp_thread = None  # type: threading.Thread
        self.event_udp = QtCore.QEvent.registerEventType()  # type: QtCore.QEvent.Type
        self.udp_numtries = 0  # type: int

        self.last_udp_message = []  # type: List[str]
        self.last_ip_address_recvd = ""
        self.isdirty = False
        self.lock = threading.Lock()
        self.check_network_thread()

    # ------------------------------------------------------------------------------
    #           add_listener
    # ------------------------------------------------------------------------------

    def add_listener(self, listenfunc: Callable):

        self.listeners.append(listenfunc)

        # ------------------------------------------------------------------------------
        #           SHOWSGE_TemperatureMonitorThread::start_udp_thread
        # ------------------------------------------------------------------------------

    def start_udp_thread(self):
        """
        Start the UDP thread that will listen on the lcoal network for Telemetry UDP packets and decode them
        """
        self.udp_thread = threading.Thread(target=self.localthread_listen_to_udp)
        self.udp_thread.start()

        # ------------------------------------------------------------------------------
        #           SHOWSGE_TemperatureMonitorThread::check_udp_thread
        # ------------------------------------------------------------------------------

    def check_network_thread(self):

        if (self.udp_on):
            if self.udp_thread is None or not self.udp_thread.is_alive():
                if (self.udp_numtries < 10):
                    self.udp_numtries += 1
                    if (self.udp_numtries >= 10):
                        print('[', self.name, ']: stopping the UDP Telemetry thread after ', self.udp_numtries, ' tries')
                        logging.warning('[%s]: stopping the UDP Telemetry thread, %d tries', self.name, self.udp_numtries)
                        self.udp_on = False
                    else:
                        self.start_udp_thread()

    # ------------------------------------------------------------------------------
    #           set_last_udp_message
    # ------------------------------------------------------------------------------

    def set_last_udp_message(self, newmessage: str, address: Tuple[str, str]):
        """Set the last telemetry message received by the system.
        This will normally run in a child thread and soes teh minimum to
        """
        self.lock.acquire()
        self.last_udp_message.append(str(newmessage))
        self.last_ip_address_recvd = address[0]
        self.isdirty = True
        self.lock.release()

    # ------------------------------------------------------------------------------
    #           decode_udp_message
    # ------------------------------------------------------------------------------

    def process_udp_message(self):
        """
        Decode the last telemtry message received from the UDP housekeeping system. This is running in the
        main application thread. The message is dispatched to all the listening callables.
        """

        self.lock.acquire()  # Block any new UDP requests until we complete this request
        #       print("process udp_message")
        if self.isdirty:  # if the new message is dirty
            for mes in self.last_udp_message:
                #               print("processing ", mes)
                for listenfunction in self.listeners:  # then
                    listenfunction(mes)  # notify all the listeners about the new message
            self.last_udp_message = []
            self.isdirty = False  # and clear the dirty flag
        self.udp_numtries = 0  # After a successful call reset the number of tries.
        #       print("finished process udp_mesaage")
        self.lock.release()

    # ------------------------------------------------------------------------------
    #           event
    # ------------------------------------------------------------------------------

    def event(self, thisevent: QtCore.QEvent) -> bool:
        """
        event handle to manage the PostEvent messages sent from the child thread after it has installed a new telemetry message.
        The event message actually comes from the main application and we must make sure the event is the one we want otherwise we just
        return back to the main event handler.

        If the event is one we need to process then we process the event which simply dispatches the message to all the objects
        listening in.

        :param thisevent:
        :return:
        """
        if (thisevent.type() == self.event_udp):
            thisevent.accept()
            self.process_udp_message()
            return True
        return super().event(thisevent)

    # ------------------------------------------------------------------------------
    #           _localthread_post_HTML_and_wait_for_return
    # ------------------------------------------------------------------------------

    def localthread_listen_to_udp(self):
        """
        This is a local thread that continuously fetches UDP telemetry packets emitted by an instrument. The packets
        are placed into the SHOWSGE_TemperatureMonitorThread::last_udp_message and a message posed to the main app. The event handler
        in SHOWSGE_TemperatureMonitorThread, which is called in the main thread, processes the message and adds it to the global housekeeping object.

        :param threadrequest:
        :param temperature_config:
        :return:
        """

        if (self.udp_on):
            try:
                name = self.name
                ipname = self.udp_ipaddress
                # ipname = '0.0.0.0'
                print('Binding UDP inbound connection for ', name, ' to ', ipname, ' on port ', self.udp_port_gse)
                serverSocket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)                 # Create a UDP socket, Notice the use of SOCK_DGRAM for UDP packets
                serverSocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)              # serverSocket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST,  1)
                try:
                    serverSocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
                except Exception:
                    pass

                serverSocket.bind((ipname, int(self.udp_port_gse)))                      # Assign IP address and port number to socket
                timeout = self.timeout
                maxmessage = self.udp_max_message

                while self.udp_on:
                    #                   print("Waiting for udp message for ", name)
                    if (timeout > 0):
                        serverSocket.settimeout(timeout / 1000.0)
                    try:
                        # print('Waiting for UDP message for module {} at {:s}:{:4d}'.format(name, ipname, self.udp_port_inst))
                        message, address = serverSocket.recvfrom(maxmessage)  # Receive the client packet along with the address it is coming from
                        if (message is not None):
                            messtr = str(message, encoding='UTF-8')
                            # print("[{}] = <{}>".format(self.name, messtr))
                        else:
                            print("UDP, no message received for ", name)
                            messtr = ""
                    except Exception as e:
                        # print("UDP exception for [{}] fired: ".format(name), e)
                        # logging.warning('[%s]: error listening to UDP broadcast from %s', self.name, self.udp_ipaddress)
                        messtr = ""
                        address = ["", ""]
                    self.set_last_udp_message(messtr, address)
                    event = QtCore.QEvent(self.event_udp)
                    self.app.postEvent(self, event)
            except:
                logging.warning('[%s]: error connecting UDP to ip address %s', self.name, self.udp_ipaddress)
        print('[', self.name, "] : thread has finished")
        logging.info('[%s]: thread has finished', self.name)
