from typing import List, Dict, Any, Union, Tuple
import datetime
import logging
import numpy as np
from PyQt5 import QtCore, QtGui, QtWidgets
from matplotlib.axes import Axes
from matplotlib.lines import Line2D

from ..application import ArgolGSE_QtUI_Application
from ..yaml_configuration import Yaml_Configuration
from ..widgets.sensor_variable_array import SensorVariableUI_Array
from ..widgets.sensor_variable_widgets import SensorVariableUI_Base
from .matplotlib_in_qtwidget import QMatplotlib_Widget


# ------------------------------------------------------------------------------
#           floor_minute
# ------------------------------------------------------------------------------

def floor_minute(t: datetime.datetime) -> datetime.datetime:
    tmin = datetime.datetime(t.year, t.month, t.day, hour=t.hour, minute=t.minute, second=0)
    return tmin


# ------------------------------------------------------------------------------
#           average_oneminute
# ------------------------------------------------------------------------------

def average_oneminute(utc: List[datetime.datetime],
                      val: List[float],
                      thisminute: datetime.datetime,
                      endminute: datetime.datetime,
                      idx: int) -> Tuple[datetime.datetime, float, int]:
    vsum = 0
    tsum = datetime.timedelta()
    npts = 0
    t = utc[idx]
    while (t < endminute) and (idx < len(utc)):
        dt = t - thisminute
        tsum = tsum + dt
        vsum = vsum + val[idx]
        npts = npts + 1
        idx = idx + 1
        if (idx < len(utc)):
            t = utc[idx]
    if (npts > 0):
        t = thisminute + (tsum / npts)
        v = vsum / npts
    return t, v, idx


# ------------------------------------------------------------------------------
#           average_minute_data
# ------------------------------------------------------------------------------

def average_minute_data(utclist: List[datetime.datetime],
                        vallist: List[float],
                        noaverage_window_mins: float):
    utc_unsorted = np.asarray(utclist)
    val_unsorted = np.asarray(vallist)
    indices = np.argsort(utc_unsorted)
    utc = np.ndarray.tolist(utc_unsorted[indices])
    val = np.ndarray.tolist(val_unsorted[indices])

    oneminute = datetime.timedelta(minutes=1)
    starttime = floor_minute(utc[0])
    endt = utc[-1]
    endttime = floor_minute(utc[-1]) - (noaverage_window_mins * oneminute)

    thisminute = starttime
    idx = 0
    newutc = []
    newval = []
    npts = len(utc)
    while (thisminute < endttime) and (idx < npts):
        endminute = thisminute + oneminute
        t = utc[idx]
        if (t >= thisminute) and (t < endminute):
            avt, v, idx = average_oneminute(utc, val, thisminute, endminute, idx)
            newutc.append(avt)
            newval.append(v)
        thisminute = endminute
    if (idx < npts):
        newutc = newutc + utc[idx:]
        newval = newval + val[idx:]
    return newutc, newval


# -----------------------------------------------------------------------------------------------------------------------
#           TimeSeries_Manager
# -----------------------------------------------------------------------------------------------------------------------

class TimeSeries_Manager:
    """
    A class managing a time series generated from one of the instrument sensor variables. The code only supports
    sensor variables that can be meaningfully converted to floats. The Timeseries_Manager registers a callback
    notification with the desired sensor variable which is invoked whenever the sensor variable is updated, typically
    from housekeeping.

    The time series manager acts as a repository for the sensor variable time-series, storing both the utc and sensor
    variable value.
    """

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, axes: Axes, sensorvariable: SensorVariableUI_Base, plotformat: str):

        self.axes: Axes = axes
        self.line: Line2D = None
        self.utc: List[datetime.datetime] = []
        self.values: List[float] = []
        self.plotformat: str = plotformat.strip()

        sensorvariable.add_update_callback(self.add_value)

    # ------------------------------------------------------------------------------
    #           close_plot
    # ------------------------------------------------------------------------------

    def close_plot(self):
        self.line = None

    # ------------------------------------------------------------------------------
    #           count
    # ------------------------------------------------------------------------------

    def count(self) -> int:
        return len(self.values)

    # ------------------------------------------------------------------------------
    #           timerange
    # ------------------------------------------------------------------------------

    def timerange(self) -> Tuple[datetime.datetime, datetime.datetime]:
        return (self.utc[0], self.utc[-1])

    # ------------------------------------------------------------------------------
    #           one_minute_average
    # ------------------------------------------------------------------------------

    def one_minute_average(self):
        if (len(self.utc) > 0):
            self.utc, self.values = average_minute_data(self.utc, self.values, 30)

    # ------------------------------------------------------------------------------
    #           add_value
    # ------------------------------------------------------------------------------

    def add_value(self, newvalue: str, utc: datetime.datetime = None):

        if (utc is None):
            utc = datetime.datetime.utcnow()
        try:
            v = float(newvalue)  # this will throw TypeError exception if we have a scalar
            self.values.append(v)
            self.utc.append(utc)
        except TypeError:
            logging.warning('TimeSeries_Manager, error adding value [', newvalue, '] to the time series. Probably due to an error converting it to a scalar floating point value')

    # ------------------------------------------------------------------------------
    #           set_value
    # ------------------------------------------------------------------------------

    def set_value(self, utc: List[datetime.datetime], value: List[float]):

        self.values = value
        self.utc = utc

    # ------------------------------------------------------------------------------
    #           update_plot
    # ------------------------------------------------------------------------------

    def update_plot(self, axes):

        if (len(self.utc) > 1):
            if self.line is None:
                lines = axes.plot_date(self.utc, self.values, self.plotformat)
                self.line = lines[0]
            else:
                self.line.set_xdata(self.utc)
                self.line.set_ydata(self.values)


# -----------------------------------------------------------------------------------------------------------------------
#           class Subplot_Manager:
# -----------------------------------------------------------------------------------------------------------------------

class Subplot_Manager:
    """
    Class used to manage one subplot on a matplotlib plotting surface. This code manages generation of the axes and
    y-bounds. The class contains a list of participating time series
    """

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, numplots: int,
                 plotconfig: Dict[str, Any],
                 sensorvariables: SensorVariableUI_Array,
                 plotwindow: QMatplotlib_Widget):

        self.subplotcolumn: int = plotconfig['subplotcolumn']
        self.subplotrow: int = plotconfig['subplotrow']
        self.layout: Tuple[float, float, float, float] = plotconfig['layout']
        self.axis_range: Tuple[float, float] = plotconfig['axis_range']
        variables: Dict[str, str] = plotconfig['variables']
        self.timeseries: List[TimeSeries_Manager] = []
        self.axes: Axes = None  # This opens a new plot

        self._open_new_plot(numplots, plotwindow)

        if (self.axes is not None):
            for varname, plotformat in variables.items():
                sensorvariable = sensorvariables.find(varname)
                if sensorvariable is not None:
                    self.timeseries.append(TimeSeries_Manager(self.axes, sensorvariable, plotformat))
                else:
                    logging.warning('Subplot_Manager, could not find sensor variable [', varname, ']. Dicarding this entry from the subplot')
        else:
            logging.warning('Subplot_Manager, error creating new axes object for the subplot. Thats not good')

    # ------------------------------------------------------------------------------
    #           open_new_plot
    # ------------------------------------------------------------------------------

    def _open_new_plot(self, numplots: int, plotwindow: QMatplotlib_Widget):

        self.axes = plotwindow.create_axes(numplots, self.subplotcolumn, self.subplotrow, exactpos=self.layout)
        self.axes.set_ylim(self.axis_range[0], self.axis_range[1])

    # ------------------------------------------------------------------------------
    #           close_plot
    # ------------------------------------------------------------------------------

    def close_plot(self):

        for p in self.timeseries:
            p.close_plot()

        self.axes.cla()

    # ------------------------------------------------------------------------------
    #           update_plot
    # ------------------------------------------------------------------------------

    def update_plot(self):

        valid = False
        mint = datetime.datetime.utcnow() + datetime.timedelta(days=10000)
        maxt = datetime.datetime.utcnow() - datetime.timedelta(days=10000)
        for series in self.timeseries:
            series.update_plot(self.axes)
            if (series.count() > 1):
                valid = True
                timerange = series.timerange()
                if (timerange[0] < mint):
                    mint = timerange[0]
                if (timerange[1] > maxt):
                    maxt = timerange[1]
        if (valid):
            self.axes.set_xlim([mint, maxt])
        self.axes.set_ylim(self.axis_range[0], self.axis_range[1])

    # ------------------------------------------------------------------------------
    #           compress_plot
    # ------------------------------------------------------------------------------

    def compress_plot(self):
        for series in self.timeseries:
            series.one_minute_average()


# -----------------------------------------------------------------------------------------------------------------------
#           PlotWindow_Manager
# -----------------------------------------------------------------------------------------------------------------------

class PlotWindow_Manager:
    """
    Class manages plotting in one window of a QT QWidget. This class manages both the QT widget and the matplotlib
    environment used to draw plots. This class manges several sub-plots defined in a top-level entry of configuration file, 
    `sensorplots.yaml`.  The management of each subplot is delegated to instances of `Subplot_Manager` which in turn 
    delegates each time series to `TimeSeries_Manager`
    """

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, qtapp: ArgolGSE_QtUI_Application,
                 sensorvariables: SensorVariableUI_Array,
                 settings: Dict[str, Any]
                 ):

        plots: List[Dict[str, Any]] = settings['plots']
        widgetname: str = settings['widget']
        interval: int = settings['refreshtime']

        self.plots: List[Subplot_Manager] = []
        self.loop_timer: QtCore.QTimer = QtCore.QTimer(qtapp)
        self.plotwindow: QMatplotlib_Widget = QMatplotlib_Widget(getattr(qtapp.ui, widgetname))

        numplots = len(plots)
        for plotcfg in plots:
            plotmanager = Subplot_Manager(numplots, plotcfg, sensorvariables, self.plotwindow)
            self.plots.append(plotmanager)

        self.loop_timer.setSingleShot(False)
        self.loop_timer.setInterval(int(interval))
        self.loop_timer.setTimerType(QtCore.Qt.CoarseTimer)
        self.loop_timer.timeout.connect(self.update_plots)
        self.loop_timer.start(int(interval))

    # ------------------------------------------------------------------------------
    #           update_plot
    # ------------------------------------------------------------------------------

    def update_plots(self):

        for p in self.plots:
            p.update_plot()
        self.plotwindow.canvas.draw()

    # ------------------------------------------------------------------------------
    #           update_plot
    # ------------------------------------------------------------------------------

    def compress_plots(self):

        for p in self.plots:
            p.compress_plot()
        self.update_plots()

    # ------------------------------------------------------------------------------
    #           close_plot
    # ------------------------------------------------------------------------------

    def close_plot(self):

        for p in self.plots:
            p.close_plot()
        self.plotwindow.close_plot()


# -----------------------------------------------------------------------------------------------------------------------
#           PlotWindows_Array
# -----------------------------------------------------------------------------------------------------------------------

class PlotWindows_Array:

    def __init__(self, qtapp: ArgolGSE_QtUI_Application,
                 sensorvariables: SensorVariableUI_Array):
        self.windows: Dict[str, PlotWindow_Manager] = {}

        conf = Yaml_Configuration('sensorplots.yaml')
        for name, cfg in conf.configuration.items():
            self.windows[name] = PlotWindow_Manager(qtapp, sensorvariables, cfg)
