from typing import Dict, Any, Tuple, List, Union, Callable
import datetime
import math
import os.path
import yaml
import numpy as np
import matplotlib.axes
from PyQt5 import QtCore, QtGui, QtWidgets

from .matplotlib_in_qtwidget import QMatplotlib_Widget


# from .showtemps_threadedudp import SHOWSGE_TemperatureMonitorThread

# ------------------------------------------------------------------------------
#           SHOWGSE_TimeSeries
# ------------------------------------------------------------------------------

class SensorVariable_TimeSeries:

    def __init__(self, plot_index: int, plotformat: str, mingoodval=-150, maxgoodval=150):
        self.utc: List[datetime.datetime] = []
        self.values: List[float] = []
        self.mingoodval: float = mingoodval
        self.maxgoodval: float = maxgoodval
        self.plot_index: int = plot_index  # type: int
        self.plotformat: str = plotformat  # type: str
        self.lines = None  # The lines object from the last call oto plot

    # ------------------------------------------------------------------------------
    #           add_value
    # ------------------------------------------------------------------------------

    def add_value(self, utc: Union[List[datetime.datetime], datetime.datetime], value: Union[List[float], float]):

        try:
            n = len(value)  # this will throw TypeError exception if we have a scalar
            for i in range(n):
                v = value[i]
                t = utc[i]
                if (v > self.mingoodval) and (v < self.maxgoodval):
                    self.values.append(v)
                    self.utc.append(datetime.datetime.utcfromtimestamp(t))
        except TypeError:
            if (value > self.mingoodval) and (value < self.maxgoodval):
                t = utc
                self.utc.append(datetime.datetime.utcfromtimestamp(t))
                self.values.append(value)

    # ------------------------------------------------------------------------------
    #           set_value
    # ------------------------------------------------------------------------------

    def set_value(self, utc: List[datetime.datetime], value: List[float]):

        self.values = value
        self.utc = utc

    # ------------------------------------------------------------------------------
    #               def close_plot(self):
    # ------------------------------------------------------------------------------

    def close_plot(self):
        self.lines = None

    # ------------------------------------------------------------------------------
    #           update_plot
    # ------------------------------------------------------------------------------

    def update_plot(self, axes):

        if (len(self.utc) > 0):
            if self.lines is None:
                self.lines = axes[self.plot_index][0].plot_date(self.utc, self.values, self.plotformat)
            else:
                self.lines[0].set_xdata(self.utc)
                self.lines[0].set_ydata(self.values)


# ------------------------------------------------------------------------------
#           SHOWGSE_Temperature_Variable_And_TimeSeries
# ------------------------------------------------------------------------------

class SHOWGSE_Temperature_And_TimeSeries(SHOWGSE_Temperature_Variable):

    def __init__(self, name: str,
                 min_value: Tuple[float, float],  # First value is orange zone, 2nd is red zone
                 max_value: Tuple[float, float],  # First value is orange zone, 2nd is red zone
                 widget_value: QtWidgets.QLabel,
                 widget_icon: QtWidgets.QLabel,
                 plot_index: int,
                 plot_format: str):
        self.name = name
        super().__init__(min_value, max_value, widget_value, widget_icon)
        self.timeseries = SHOWGSE_TimeSeries(plot_index, plot_format)

    # ------------------------------------------------------------------------------
    #           set_value
    # ------------------------------------------------------------------------------

    def set_value(self, utc: List[datetime.datetime], value: List[float]):
        super().set_value(value[-1])
        self.timeseries.set_value(utc, value)


# ------------------------------------------------------------------------------
#           Class SHOWGSE_TemperatureGui
# ------------------------------------------------------------------------------

class SHOWGSE_TemperatureGui:
    """
    Class that maps GUI based widgets to GSE temperature variables. 
    """

    def __init__(self, globalobject: 'mainglobalobject.SHOWGSE_MainGlobalObject'):

        self.globalobject = globalobject  # type: 'mainglobalobject.SHOWGSE_MainGlobalObject'
        ui = globalobject.gui_app.ui
        self.gui_temperatures = {}  # type: Dict[str,SHOWGSE_Temperature_And_TimeSeries]
        self.telemetry = globalobject.hktemps.hktemps
        self.PLOT_TEMPERATURES = ui.PLOT_TEMPERATURES
        self.plot_counter = 1
        self.open_new_plot()

        # ---- Initialize the temperature widgets

        self.gui_temperatures["FPA"] = SHOWGSE_Temperature_And_TimeSeries("FPA", [-0.7, -2.0], [16.0,
                                                                                                25.0], ui.TEMP_FPA, ui.TEMP_FPA_STATUS, 0, 'ko-')

        self.gui_temperatures["TOPSHS"] = SHOWGSE_Temperature_And_TimeSeries("Top SHS", [18.0, 15.0], [25.0,
                                                                                                       34.0], ui.TEMP_TOPSHS, ui.TEMP_TOPSHS_STATUS, 1, 'ko-')
        self.gui_temperatures["BOTSHS"] = SHOWGSE_Temperature_And_TimeSeries("Bot SHS", [18.0, 15.0], [25.0,
                                                                                                       24.0], ui.TEMP_BOTSHS, ui.TEMP_BOTSHS_STATUS, 1, 'ro-')
        self.gui_temperatures["OPTOBOX"] = SHOWGSE_Temperature_And_TimeSeries("Opto Box", [18.0, 15.0], [25.0,
                                                                                                         24.0], ui.TEMP_OPTOBOX, ui.TEMP_OPTOBOX_STATUS, 1, 'go-')

        self.gui_temperatures["TOPHEATER"] = SHOWGSE_Temperature_And_TimeSeries("Top Heater", [-0.1, -0.2], [90.0,
                                                                                                             100.0], ui.TOP_HEATER, ui.TOP_HEATER_STATUS, 2, 'ko-')
        self.gui_temperatures["BOTHEATER"] = SHOWGSE_Temperature_And_TimeSeries("Bot Heater", [-0.1, -0.2], [90.0,
                                                                                                             100.0], ui.BOTTOM_HEATER, ui.BOTTOM_HEATER_STATUS, 2, 'ro-')

        self.gui_temperatures["EBOX"] = SHOWGSE_Temperature_And_TimeSeries("E-Box", [15.0, 10.0], [40.0,
                                                                                                   60.0], ui.TEMP_EBOX, ui.TEMP_EBOX_STATUS, 3, 'ko-')
        self.gui_temperatures["CAMERACASE"] = SHOWGSE_Temperature_And_TimeSeries("Camera Case", [18.0, 15.0], [37.0,
                                                                                                               39.0], ui.TEMP_CASE, ui.TEMP_CASE_STATUS, 3, 'go-')
        self.gui_temperatures["Q7"] = SHOWGSE_Temperature_And_TimeSeries("Q7", [15.0, 10.0], [70.0,
                                                                                              90.0], ui.TEMP_Q7, ui.TEMP_Q7_STATUS, 3, 'ro-')
        self.gui_temperatures["PCB"] = SHOWGSE_Temperature_And_TimeSeries("PCB", [18.0, 15.0], [37.0,
                                                                                                39.0], ui.TEMP_PCB, ui.TEMP_PCB_STATUS, 3, 'bo-')

        self.gui_temperatures["WINGPOD"] = SHOWGSE_Temperature_And_TimeSeries("Wingpod", [-50.0, -60.0], [28.0,
                                                                                                          25.0], ui.TEMP_WINGPOD, ui.TEMP_WINGPOD_STATUS, 4, 'ko-')
        self.gui_temperatures["PERISCOPE"] = SHOWGSE_Temperature_And_TimeSeries("Periscope", [-50.0, -60.0], [28.0,
                                                                                                              25.0], ui.TEMP_PERISCOPE, ui.TEMP_PERISCOPE_STATUS, 4, 'ro-')
        self.utcwidget = ui.TEMP_UTC_STR

        globalobject.hktemps.add_listener(self.update_current_values_from_latest_telemetry)

    #      ui.BUTTON_CMD_COLLECT_IMAGES.clicked.connect(self.print_diagnostics)
    #     ui.BUTTON_CMD_APPLY_SETTINGS.clicked.connect(self.ButtonClick_ApplySettings)
    # self.temperature_threads = SHOWSGE_TemperatureMonitorThread(self.globalobject.gui_app, globalobject, self, self.telemetry,  self.globalobject.gse_config["TEMPERATURES"], self.globalobject.gse_config["DEFAULTS"] )

    # ------------------------------------------------------------------------------
    #           set_html_commandstate_callback
    # ------------------------------------------------------------------------------

    # def set_html_commandstate_callback(self, callback : Callable):
    #    self.temperature_threads.set_html_commandstate_callback( callback )

    # ------------------------------------------------------------------------------
    #           update_current_values_from_latest_telemetry
    # ------------------------------------------------------------------------------

    def update_current_values_from_latest_telemetry(self, telemetry: OWL640Telemetry):
        """ Update the GUI using the latest values in the Telemetry Housekeeping array"""
        if (len(telemetry.utc) > 0):
            utc = []
            for t in telemetry.utc:
                utc.append(datetime.datetime.utcfromtimestamp(t))

            self.gui_temperatures["FPA"].set_value(utc, telemetry.sensor_temp)
            self.gui_temperatures["TOPSHS"].set_value(utc, telemetry.top_temperature)
            self.gui_temperatures["BOTSHS"].set_value(utc, telemetry.bot_temperature)
            self.gui_temperatures["OPTOBOX"].set_value(utc, telemetry.optobox_temperature)
            self.gui_temperatures["CAMERACASE"].set_value(utc, telemetry.cameracase_temp)
            self.gui_temperatures["TOPHEATER"].set_value(utc, telemetry.top_heater_power)
            self.gui_temperatures["BOTHEATER"].set_value(utc, telemetry.bot_heater_power)
            self.gui_temperatures["EBOX"].set_value(utc, telemetry.ebox_temperature)
            self.gui_temperatures["Q7"].set_value(utc, telemetry.q7_temp)
            self.gui_temperatures["PCB"].set_value(utc, telemetry.sensor_pcb_temp)
            self.gui_temperatures["WINGPOD"].set_value(utc, telemetry.wingpod_window_temp)
            self.gui_temperatures["PERISCOPE"].set_value(utc, telemetry.periscope_temp)
            #            utc = utc[-1]
            utc = datetime.datetime.utcnow()
            strval = utc.isoformat(timespec='milliseconds')
            self.utcwidget.setText(strval)

    # ------------------------------------------------------------------------------
    #           open_new_plot
    # ------------------------------------------------------------------------------

    def open_new_plot(self):
        self.plotwindow = QMatplotlib_Widget(self.PLOT_TEMPERATURES)
        self.axes = []
        self.axes.append([self.plotwindow.create_axes(5, 1, 1, exactpos=[0.05, 0.83, 0.90, 0.16]), [-10, 30]])
        self.axes.append([self.plotwindow.create_axes(5, 1, 2, exactpos=[0.05, 0.63, 0.90, 0.16]), [15, 30]])
        self.axes.append([self.plotwindow.create_axes(5, 1, 3, exactpos=[0.05, 0.43, 0.90, 0.16]), [0, 100]])
        self.axes.append([self.plotwindow.create_axes(5, 1, 4, exactpos=[0.05, 0.23, 0.90, 0.16]), [0, 100]])
        self.axes.append([self.plotwindow.create_axes(5, 1, 5, exactpos=[0.05, 0.03, 0.90, 0.16]), [-100, 50]])

    # ------------------------------------------------------------------------------
    #           close_plot
    # ------------------------------------------------------------------------------

    def close_plot(self):
        for key, p in self.gui_temperatures.items():
            p.timeseries.close_plot()

        for ax in self.axes:
            ax[0].cla()
            ax[0] = None

        self.plotwindow.close_plot()
        self.plotwindow = None

    # ------------------------------------------------------------------------------
    #           update_plots
    # ------------------------------------------------------------------------------

    def update_plots(self):

        #        if ( self.plot_counter > 20):
        #            self.close_plot()
        #            self.plot_counter = 0
        #        else:
        # if (self.plot_counter   == 1): self.open_new_plot()

        if (len(self.telemetry.utc) > 1):
            self.plot_counter += 1

            for key, p in self.gui_temperatures.items():
                p.timeseries.update_plot(self.axes)

            maxt = datetime.datetime.utcfromtimestamp(self.telemetry.utc[-1])
            mint = datetime.datetime.utcfromtimestamp(self.telemetry.utc[0])
            for ax in self.axes:
                limits = ax[0].get_xlim()
                ax[0].set_xlim([mint, maxt])
                limits = ax[1]
                ax[0].set_ylim(limits[0], limits[1])

            self.plotwindow.canvas.draw()

    # ------------------------------------------------------------------------------
    #           print_diagnostics
    # ------------------------------------------------------------------------------
    @QtCore.pyqtSlot()
    def print_diagnostics(self):
        #

        fullname = __file__
        [basedir, spare] = os.path.split(fullname)  # split out the list_of_files
        [basedir, spare] = os.path.split(basedir)
        fullname = basedir + os.sep + 'testdata' + os.sep + 'telemetry' + os.sep + '2017-04-05_20-28'

        npts = self.telemetry.readtelemetryfolders(fullname)
        tm = self.telemetry

        n = len(tm.utc)
        t0 = datetime.datetime.utcnow() - datetime.timedelta(seconds=n)
        for i in range(n):
            t = t0 + datetime.timedelta(seconds=i)
            tm.utc[i] = t
            tm.utc_array[i] = t.timestamp()

        self.update_current_values_from_latest_telemetry(tm)
