from typing import Dict, Any, List, Callable  # , Deque
from collections import deque
import threading
import os
import pysftp
import time
from PyQt5 import QtCore
from .yaml_configuration import Yaml_Configuration
from .application import ArgolGSE_QtUI_Application


# ------------------------------------------------------------------------------
#           queued_ftp_download_entry
# ------------------------------------------------------------------------------

class queued_ftp_download_entry:

    def __init__(self, downloadtype: str = None,
                 remote_filename: str = None,
                 local_fullname: str = None,
                 download_complete_callback: Callable = None):
        self.downloadtype: str = downloadtype  # Can be NOne. Otehrwise string identifying special download type using special Logger messages, eg 'ARCHIVE' or 'SAMPLED'
        self.remote_filename: str = remote_filename.strip()  # Full absolute path to the remote filename
        self.local_fullname: str = local_fullname.strip()  # The local location to store the file
        self.numtries: int = 0  # the number of tries. we abort after 5 attempts
        self.download_complete_callback: Callable = download_complete_callback  # We can call this guy with the entry once downloaded is completed if required


# ------------------------------------------------------------------------------
#           class sftp:
# ------------------------------------------------------------------------------

class sftp:

    def __init__(self, host_ip_address: str = None) -> None:

        network_settings = Yaml_Configuration('network.yaml').configuration['FTP']
        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None
        if (host_ip_address is None) or (len(host_ip_address) == 0):
            host_ip_address = network_settings["ftp_ip"]
        username = network_settings["username"]
        password = network_settings["password"]

        try:
            print('Queued FTP: Attempting to make sftp connection to firmware at ', host_ip_address)
            self.connection = pysftp.Connection(host=host_ip_address,
                                                username=username,
                                                password=password,
                                                default_path='/',
                                                cnopts=cnopts)
            print('Queued FTP: Connection made to ', host_ip_address)
        except Exception as err:
            print('Queued FTP. Error creating an instance of secure FTP client. Error = [', str(err), "]")
            self.connection = None  # Assign the error to the connection, can check if good connection with isinstance


# ------------------------------------------------------------------------------
#           cats_ftp
# ------------------------------------------------------------------------------

class Argol_Queued_FTP(QtCore.QObject):

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, qtapp, ftp_address_callback: Callable = None):

        super().__init__()
        self.app: ArgolGSE_QtUI_Application = qtapp
        self.ftp_address_callback: Callable = ftp_address_callback  # A callable function that returns the IP Address of the server. This information is usually derived from the HK UDP connection
        self.filelist = deque()  #: Deque[queued_ftp_download_entry] = deque()           # The list of files waiting to be downloaded, Deque Typing not avilable until 3.6.1
        self.quit: bool = False  # Flags that the FTP thread can quit
        self.conditionvariable: threading.Condition = threading.Condition()  # Condition Variable used to control access to FTP threaded loop
        self.thread: threading.Thread = threading.Thread(target=self._process_files)  # Thread frunning the FTP download loop
        self.ftp: sftp = None  # The FTP object. It is not fired up until needed.
        self.numftptries: int = 0  # The number of tries attempting to make an FTP connection
        self._download_complete_entry: List[
            queued_ftp_download_entry] = []  # Stores the list of download events completed
        self._download_complete_lock: threading.Lock = threading.Lock()
        self.event_download_complete: QtCore.QEvent.Type = QtCore.QEvent.registerEventType()
        self.thread.start()  # Start the ftp download thread.

    # ------------------------------------------------------------------------------
    #           stop
    # ------------------------------------------------------------------------------

    def stop(self):
        """
        Stop the FTP download thread. The FTP download thread will finish any download in
        progress, if any, and then the thread will quit.
        """

        self.quit = True
        self.conditionvariable.notify_all()

    # ------------------------------------------------------------------------------
    #           queue_file
    # ------------------------------------------------------------------------------

    def queue_file(self, entry: queued_ftp_download_entry) -> bool:
        """
        Queue a file for download. The caller must fill out the entire entry structure
        before submitting this job. This is called from the users thread. NOte that the
        callback, if provided, will occur in the download thread. It should be used to send QEvents
        to the QApplication as pyQt code may not be thread safe.
        """

        ok = (entry.local_fullname is not None) and os.path.exists(entry.local_fullname)  # dont bother if the local file exists
        if not ok:
            self.conditionvariable.acquire()
            ok = (len(self.filelist) < 15)
            if (ok):
                self.filelist.append(entry)
                self.conditionvariable.notify_all()
            self.conditionvariable.release()
            if (not ok):
                print("FTP: remote file ", entry.remote_filename, " was not queues for download as the queue is full")
        return ok

    # ------------------------------------------------------------------------------
    #           download_entry
    # ------------------------------------------------------------------------------

    def _download_entry(self, entry: queued_ftp_download_entry):
        """
        Downloads the given entry. This is normally running within the FTP download thread.
        It executes a callback to a user supplied callable. This callback function should try and restrict
        what it does with pyQT objects. It is recommended that a custom QEvent is sent to the main application
        so the Event can be processed by the main pyQT message processing loop
        """

        entry.numtries += 1
        if (self.ftp is None):
            if (self.numftptries < 10):
                host_ip_address = None
                if (
                        self.ftp_address_callback is not None):  # If we have a callback functo provide the FTP IP address (often from the HK UDP packets)
                    host_ip_address = self.ftp_address_callback()  # Then get the host ip addres sfrom teh callback
                self.ftp = sftp(host_ip_address=host_ip_address)  # Create the FTP connnection
                if (self.ftp.connection is None):  # If the connection was not successful
                    self.ftp = None  # Then destroy the FTP object and try again
                    self.numftptries += 1  # dont try to connect more than 10 times.
                    time.sleep(0.5)  # and sleep for a little while

        ok = self.ftp is not None
        if (ok):
            self.numftptries = 0
            local_dir, local_name = os.path.split(entry.local_fullname)
            os.makedirs(local_dir, exist_ok=True)
            remote_fullname = entry.remote_filename
            if (remote_fullname[0] == '/'):
                remote_fullname = remote_fullname[1:]

            print('FTP, Get [%s] and stick in [%s]' % (remote_fullname, entry.local_fullname))

            # try:
            self.ftp.connection.get(remote_fullname, localpath=entry.local_fullname, preserve_mtime=True)
            print('Queued FTP: finished downloading ', remote_fullname)
            # except:
            #    ok = False
            if (ok) and (entry.download_complete_callback is not None):
                entry.download_complete_callback(entry)
        if (not ok):
            if (entry.numtries < 5):
                print("FTP FAILED, [", entry.remote_filename, "]. Requeueing attempt ", entry.numtries, " of 5")
                self.queue_file(entry)
            else:
                print("FTP FAILED [", entry.remote_filename, "]. Abandoning attempt after 5 retries")
        return ok

    # ------------------------------------------------------------------------------
    #           _process_files
    # ------------------------------------------------------------------------------

    def _process_files(self):
        """
        The main FTP download thread. It waits for a notification of a file. Chjecks the list of queued files and then fetches
        them. Note that if an object fails it is pushed onto the end of the list
        """
        print("FTP thread is firing up")
        with self.conditionvariable:
            while not self.quit:
                self.conditionvariable.acquire()
                n = len(self.filelist)
                if ((n > 0) and (not self.quit)):
                    entry = self.filelist.popleft()
                    self.conditionvariable.release()
                    self._download_entry(entry)
                else:
                    self.conditionvariable.wait()
        print("FTP thread is closing down")

    # ------------------------------------------------------------------------------
    #           event
    # ------------------------------------------------------------------------------

    def event(self, thisevent: QtCore.QEvent) -> bool:
        """
        event handle to manage the PostEvent messages sent from the child thread after it has completed an FTP download
        The event message comes from the main application and we see if the event is the one we want otherwise we
        return back to the main event handler.
        """
        if (thisevent.type() == self.event_download_complete):
            thisevent.accept()
            self.event_download_complete_process()
            return True
        return super().event(thisevent)

    # ------------------------------------------------------------------------------
    #           event_fire_download_complete
    # ------------------------------------------------------------------------------

    def event_download_complete_fire(self, ftpentry: queued_ftp_download_entry):
        """
        Posts events to the qt application so the callbacks can be executed  in the
        proper QT threading context. It is possible multiple FTP callbackcs may occur
        so the callbacks are stored in an internal list.

        The QT app will eventually call `event` in the porper QT threading context which invokes
        `event_download_complete_process`. At this point all the callbacks are popped off the list
        and invoked.
        """
        if (ftpentry.download_complete_callback is not None):
            self._download_complete_lock.acquire()
            self._download_complete_entry.append(ftpentry)
            self._download_complete_lock.release()
            event = QtCore.QEvent(self.event_download_complete)
            self.app.postEvent(self, event)

    # ------------------------------------------------------------------------------
    #           event_process_download_complete
    # ------------------------------------------------------------------------------

    def event_download_complete_process(self):
        """
        Handles the event used to process callbacks at the end of the FTP transfer. This
        function is invoked in the proper threading context of the QT application.
        """

        self._download_complete_lock.acquire()  # Acquire lock so event_download_complete_fire cannot add anyy more entries
        while len(self._download_complete_entry) > 0:  # See if we have any entries to process
            ftpentry = self._download_complete_entry.pop()  # if we do then pop off the last entry added
            self._download_complete_lock.release()  # and release the lock so event_download_complete_fire can add more entries
            ftpentry.download_complete_callback(ftpentry)  # execute the callback in the proper threading context for QT Apps
            self._download_complete_lock.acquire()  # reacquire the lock in preaparation for the 'while' check
        self._download_complete_lock.release()  # and release teh lck when don.
