from typing import Dict, Any, List
from PyQt5 import QtCore, QtGui, QtWidgets
import datetime
import re
import os.path
from ..application import ArgolGSE_MainGlobalObject

# from .decode_filelisting import SHOWGSE_FileListingEntry, SHOWGSE_LoadFileListings
# from .showgse_displayimage import SHOWGSE_ImageDisplayGui

g_icon_fullfolder = QtGui.QIcon(":/icons/resources/icons/folder.ico")
g_icon_halffullfolder = QtGui.QIcon(":/icons/resources/icons/halffullfolder.ico")
g_icon_emptyfolder = QtGui.QIcon(":/icons/resources/icons/emptyfolder.ico")
g_icon_fullfile = QtGui.QIcon(":/icons/resources/icons/fullfile.ico")
g_icon_emptyfile = QtGui.QIcon(":/icons/resources/icons/emptyfile.ico")


# ------------------------------------------------------------------------------
#           decode_dirnamefromshow
#   decode from yyyy-mm-dd_hh-mm to yyyy-mmm-dd hh:mm
# ------------------------------------------------------------------------------

def decode_dirname_from_show(showdirname: str) -> str:
    m = re.match('([0-9]+)-([0-9]+)-([0-9]+)_([0-9]+)-([0-9]+)$', showdirname)
    if (m is not None):
        year = int(m.group(1))
        month = int(m.group(2))
        day = int(m.group(3))
        hour = int(m.group(4))
        mins = int(m.group(5))
        dirstr = '%04i-%02i-%02i %02i:%02i' % (year, month, day, hour, mins)
    else:
        dirstr = None
    return dirstr


# ------------------------------------------------------------------------------
#           encode_treeview_dirname_to_show
# ------------------------------------------------------------------------------

def encode_treeview_dirname_to_show(treeview_dirname):
    m = re.match('([0-9]+)-([0-9]+)-([0-9]+) ([0-9]+):([0-9]+)$', treeview_dirname)
    year = int(m.group(1))
    month = int(m.group(2))
    day = int(m.group(3))
    hour = int(m.group(4))
    mins = int(m.group(5))
    showstr = '%04i-%02i-%02i_%02i-%02i' % (year, month, day, hour, mins)
    return showstr


# ------------------------------------------------------------------------------
#           encode_treeview_dirname_to_utc
# ------------------------------------------------------------------------------

def encode_treeview_dirname_to_utc(treeview_dirname) -> datetime.datetime:
    m = re.match('([0-9]+)-([0-9]+)-([0-9]+) ([0-9]+):([0-9]+)$', treeview_dirname)
    year = int(m.group(1))
    month = int(m.group(2))
    day = int(m.group(3))
    hour = int(m.group(4))
    mins = int(m.group(5))
    utc = datetime.datetime(year, month, day, hour, mins, 0)
    return utc


# ------------------------------------------------------------------------------
#           ArgolGSE_FileListing_TreeView_ImageEntry
# ------------------------------------------------------------------------------

class ArgolGSE_FileListing_TreeView_ImageEntry:

    def __init__(self, utcstr: str, value: SHOWGSE_FileListingEntry):

        self.treeviewname = utcstr
        self.value = value  # Type: SHOWGSE_FileListingEntry, details about the image received from SHOW instrument
        self.is_local = False  # Type: bool, True if the file is stored locally
        self.local_filename = 'N/A'

    # ------------------------------------------------------------------------------
    #           status_icon
    # ------------------------------------------------------------------------------

    def status_icon(self):

        global g_icon_fullfile
        global g_icon_emptyfile

        if self.is_local:
            icon = g_icon_fullfile
        else:
            icon = g_icon_emptyfile
        return icon

    # ------------------------------------------------------------------------------
    #           set_local_filename
    # ------------------------------------------------------------------------------

    def set_local_filename(self, filename):

        if filename is not None and os.path.exists(filename):
            self.is_local = True  # Type: bool, True if the file is stored locally
            self.local_filename = filename
        else:
            self.is_local = False  # Type: bool, True if the file is stored locally
            self.local_filename = "N/A"


# ------------------------------------------------------------------------------
#           ArgolGSE_FileListing_TreeView_DirectoryEntry
# ------------------------------------------------------------------------------

class ArgolGSE_FileListing_TreeView_DirectoryEntry:

    def __init__(self, treeview_dirname: str):

        oneminute = datetime.timedelta(minutes=1, seconds=10)
        self.utc_start = encode_treeview_dirname_to_utc(treeview_dirname)
        self.utc_end = self.utc_start + oneminute  # Type: datetime.datetime
        self.dirname = treeview_dirname
        self.list_of_files = {}  # Type: Dict[str, ArgolGSE_FileListing_TreeView_ImageEntry], the string is the string displayed in the tree view
        self.load_status = 0  # 0 = not loaded, 1 = partially loaded, 2 = fully loaded (all based upon UTC)

    # ------------------------------------------------------------------------------
    #           is_fully_loaded
    # ------------------------------------------------------------------------------

    def is_fully_loaded(self):
        return (self.load_status == 2)

    # ------------------------------------------------------------------------------
    #           update_load_status
    # ------------------------------------------------------------------------------

    def update_load_status(self):
        """
        Update the 
        :return: 
        """
        utc = datetime.datetime.utcnow()
        if (utc > self.utc_end):
            self.load_status = 2  # we are well beyond the end time of the folder so the last listing must be complete
        else:
            self.load_status = 1  # otherwise the last listing is potentially partial

    # ------------------------------------------------------------------------------
    #           status_icon
    # ------------------------------------------------------------------------------

    def status_icon(self):

        global g_icon_fullfolder
        global g_icon_halffullfolder
        global g_icon_emptyfolder

        if self.load_status == 2:
            icon = g_icon_fullfolder
        elif self.load_status == 1:
            icon = g_icon_halffullfolder
        else:
            icon = g_icon_emptyfolder
        return icon

    # ------------------------------------------------------------------------------
    #           find_entry
    # ------------------------------------------------------------------------------

    def find_entry(self, utcstr: str, value: SHOWGSE_FileListingEntry):

        entry = self.list_of_files.get(utcstr)
        if (entry is None):
            self.list_of_files[utcstr] = ArgolGSE_FileListing_TreeView_ImageEntry(utcstr, value)
            entry = self.list_of_files[utcstr]
        return entry

    # ------------------------------------------------------------------------------
    #           add_files
    # ------------------------------------------------------------------------------

    def add_files(self, filelisting: Dict[str, SHOWGSE_FileListingEntry], treeview: QtWidgets.QTreeWidget):

        treeitems = treeview.findItems(self.dirname, QtCore.Qt.MatchFixedString)
        directoryitem = treeitems[0]  # we must find this directory entry in our tree

        for key, v in filelisting.items():
            utcstr = v.utc.isoformat(sep=' ', timespec='milliseconds')
            entry = self.find_entry(utcstr, v)
            treeitems = treeview.findItems(utcstr, QtCore.Qt.MatchFixedString)
            if len(treeitems) > 0:
                item = treeitems[0]
            else:
                exptime = str(int(v.exposure_time))
                item = QtWidgets.QTreeWidgetItem(directoryitem, [utcstr, exptime,
                                                                 v.science_mode], QtWidgets.QTreeWidgetItem.Type)
                item.setIcon(0, entry.status_icon())

        self.update_load_status()
        directoryitem.setIcon(0, self.status_icon())


# ------------------------------------------------------------------------------
#           ArgolGSE_FileListingGui
# ------------------------------------------------------------------------------

class ArgolGSE_FileListingGui:
    """
    """

    def __init__(self, globalobject: ArgolGSE_MainGlobalObject, treeview_widget_name: str):
        self.globalobject = globalobject  # type: 'mainglobalobject.GSE_MainGlobalObject'
        self.show_file_lists = SHOWGSE_LoadFileListings(globalobject)  # object that fetches file listings
        self.directories = {}
        ui = globalobject.gui_app().ui

        self.imagedisplay_gui = SHOWGSE_ImageDisplayGui(globalobject)

        self.lightgraybrush = QtGui.QBrush(QtCore.Qt.lightGray)
        self.darkgraybrush = QtGui.QBrush(QtCore.Qt.darkGray)
        self.treeview = getattr(ui, treeview_widget_name)
        self.timer = QtCore.QTimer(globalobject.gui_app)
        self.timer.timeout.connect(self.twosecondtimer)
        self.timer.start(2000)
        self.lasttimertime = (datetime.datetime.utcnow()).replace(second=0, microsecond=0)

        self.configure_tree_view()  # globalobject.gui_app)
        self.treeview.itemPressed.connect(self.item_pressed)

    # ------------------------------------------------------------------------------
    #           configure_tree_view
    # ------------------------------------------------------------------------------

    def configure_tree_view(self):

        # self.treeview.setColumnCount(3)
        self.treeview.setHeaderLabels(['UTC', 'Exp Time', 'Science Mode'])
        self.treeview.setColumnWidth(0, 200)
        self.treeview.setColumnWidth(1, 50)
        self.treeview.setColumnWidth(2, 120)

    # ------------------------------------------------------------------------------
    #           directory_entry
    # ------------------------------------------------------------------------------

    def directory_entry(self, showdirname: str, isshowdir=True) -> ArgolGSE_FileListing_TreeView_DirectoryEntry:

        entry = None
        if (isshowdir):
            treeview_dirstr = decode_dirname_from_show(showdirname)
        else:
            treeview_dirstr = showdirname

        if (treeview_dirstr is not None):
            entry = self.directories.get(treeview_dirstr)
            if entry is None:
                self.directories[treeview_dirstr] = ArgolGSE_FileListing_TreeView_DirectoryEntry(treeview_dirstr)
                entry = self.directories[treeview_dirstr]
        return entry

    # ------------------------------------------------------------------------------
    #           directory_item
    # ------------------------------------------------------------------------------

    def directory_item(self, entry: ArgolGSE_FileListing_TreeView_DirectoryEntry) -> QtWidgets.QTreeWidgetItem:

        treeitems = self.treeview.findItems(entry.dirname, QtCore.Qt.MatchFixedString)
        if (len(treeitems) > 0):
            item = treeitems[0]
        else:
            item = QtWidgets.QTreeWidgetItem([entry.dirname, '', ''], QtWidgets.QTreeWidgetItem.Type)
            item.setChildIndicatorPolicy(QtWidgets.QTreeWidgetItem.DontShowIndicatorWhenChildless)
            item.setIcon(0, entry.status_icon())
            self.treeview.addTopLevelItem(item)
        return item

    # ------------------------------------------------------------------------------
    #           process_show_list_of_dirs
    # ------------------------------------------------------------------------------

    def process_list_of_dirs(self, dirlisting: List[str]):
        """
        This is the method called to process the full list of imaging directories received from the SHW instrument.
        It is invoked by the action ButtonClick_RefreshDirectory
        
        :param dirlisting: 
        :return: 
        """

        for v in dirlisting:
            entry = self.directory_entry(v)
            if (entry is not None):
                item = self.directory_item(entry)

    # ------------------------------------------------------------------------------
    #           process_show_filelisting
    # ------------------------------------------------------------------------------

    def process_show_filelisting(self, dirname: str, filelisting: Dict[float, SHOWGSE_FileListingEntry]):

        entry = self.directory_entry(dirname)
        if (entry is not None):
            entry.add_files(filelisting, self.treeview)

    # ------------------------------------------------------------------------------
    #           start_download_and_display_image
    # ------------------------------------------------------------------------------

    def start_download_and_display_image(self, item: QtWidgets.QTreeWidgetItem):

        treeview_dirname = item.parent().text(0)
        showdirname = encode_treeview_dirname_to_show(treeview_dirname)
        utcstr = item.text(0)
        entry = self.directory_entry(treeview_dirname, isshowdir=False)
        fileentry = entry.find_entry(utcstr, None)
        filename = showdirname + '/' + fileentry.value.name
        self.show_file_lists.fetch_image(filename, treeview_dirname, utcstr, item, self.finish_image_download_and_display)

    # ------------------------------------------------------------------------------
    #           finish_image_download_and_display
    # ------------------------------------------------------------------------------

    def finish_image_download_and_display(self, filename: str, treeview_dirname: str, utcstr: str, item: QtWidgets.QTreeWidgetItem):

        entry = self.directory_entry(treeview_dirname, isshowdir=False)
        fileentry = entry.find_entry(utcstr, None)
        fileentry.set_local_filename(filename)
        item.setIcon(0, fileentry.status_icon())
        self.imagedisplay_gui.update_image(filename, utcstr)

    # ------------------------------------------------------------------------------
    #           twosecondtimer
    # ------------------------------------------------------------------------------

    @QtCore.pyqtSlot()
    def twosecondtimer(self):
        """
        Every two seconds check to see if a new minute has occured. It it has and we are more than 3 seconds into the minute then
        assume SHOW has created a new folder and add the folder to the list of directories. This avoids users having to constantly refresh
        :return:
        """
        tnow = datetime.datetime.utcnow()
        tminute = tnow.replace(second=0, microsecond=0)
        if (tminute > self.lasttimertime) and (tnow.second > 3):
            self.lasttimertime = tminute
            showdirname = '%04i-%02i-%02i_%02i-%02i' % (tminute.year, tminute.month, tminute.day, tminute.hour, tminute.minute)
            self.process_list_of_dirs([showdirname])

    # ------------------------------------------------------------------------------
    #           item_pressed
    # #------------------------------------------------------------------------------

    @QtCore.pyqtSlot(QtWidgets.QTreeWidgetItem, int)
    def item_pressed(self, item: QtWidgets.QTreeWidgetItem, column: int):

        text = item.text(0)
        isfile = text.find('.') >= 0  # Directory listings dont have a decimal point. Image listings do
        if (isfile):
            self.start_download_and_display_image(item)
        else:
            dirname = encode_treeview_dirname_to_show(text)
            self.show_file_lists.fetch_list_of_files(dirname, self.process_show_filelisting)
