from typing import Any, Sequence, Dict, List, Union, NamedTuple, Tuple, TypeVar, Callable
from PyQt5 import QtCore, QtGui, QtWidgets
import math
from lxml import html
import time
import datetime
import re
import threading
import os
import os.path

# from ..mainglobalobject import GSE_MainGlobalObject
from ..htmldecode.threadedhtml import ThreadedHTMLRequest

SHOWGSE_FileListingEntry = NamedTuple('SHOWGSE_FileListingEntry', [('name', str), ('utc', datetime.datetime),
                                                                   ('exposure_time', float), ('science_mode', str),
                                                                   ('link', str)])


# ------------------------------------------------------------------------------
#           SHOWGSE_LoadFileListings
# ------------------------------------------------------------------------------

class SHOWGSE_LoadFileListings:

    def __init__(self, mainglobalobject: 'GSE_MainGlobalObject'):
        tnow = datetime.date.today()
        datestr = tnow.strftime('%Y-%m-%d')

        self.globalobject = mainglobalobject
        self.output_folder = mainglobalobject.gse_config["DEFAULTS"]["base_output_dir"] + os.sep + datestr
        self.html_config = mainglobalobject.gse_config["show_html"]
        self.threadrequest = None
        self.lock = threading.Lock()
        self.isbusy = False
        self.callback = None  # Type: Callable
        self.dirname = None

    # ------------------------------------------------------------------------------
    #           decode_mainpage
    # ------------------------------------------------------------------------------

    def decode_filelisting(self, content: str):
        """
        Decodes the information from the mainpage.py generated by the SHOW instrument.
        The fields are decoded into 1) HK temperatures, 2) Miscellaneous and 3) Command Status
        fields. These sections can then be given to the current instrument state and checked against the
        expected values.
    
        :param tree: The HTML object generated from parsing the HTML retuned from SHOW mainpage.py  
        :return: A tuple of three dictionaries [ HK Temps, Misccellaneous, Command Status]
        """

        dirlisting = []
        filelisting = {}
        if (content is not None) and (len(content) > 0):
            tree = html.fromstring(content)
            td = tree.findall(r'.//td')
            dirlisting = []
            filelisting = {}
            if (td is not None):
                n = len(td)
                i = 0
                while (i < n):
                    nametd = td[i]
                    aentry = nametd.findall(r'.//a')
                    link = aentry[0].get('href')
                    name = str(td[i].text_content())
                    date = str(td[i + 1].text_content())
                    attrib = str(td[i + 2].text_content())
                    type = str(td[i + 3].text_content())
                    if type == 'Directory':
                        if (name != 'Parent Directory/'):
                            dirlisting.append(name[0:-1])  # remove the trailing '/'
                    else:
                        m = re.match(r'([0-9\.]+)-([0-9]+)ms[-]*([0-9A-Za-z_\-]*).raw$', name)
                        if (m is not None):
                            utc = datetime.datetime.utcfromtimestamp(float(m.group(1)))
                            expt = float(m.group(2))
                            sciencemode = m.group(3)
                            entry = SHOWGSE_FileListingEntry(name=name, utc=utc, exposure_time=expt, science_mode=sciencemode, link=link)
                            filelisting[entry.utc] = entry
                    i = i + 4
        return dirlisting, filelisting

    # ------------------------------------------------------------------------------
    #           wait_for_not_busy
    # ------------------------------------------------------------------------------

    def wait_for_not_busy(self):

        busy = True
        while busy:
            self.lock.acquire()
            busy = self.isbusy
            if (not busy):
                self.isbusy = True
                self.lock.release()
            else:
                self.lock.release()
                self.globalobject.gui_app.processEvents()
                time.sleep(0.01)

    # ------------------------------------------------------------------------------
    #           is_busy
    # ------------------------------------------------------------------------------

    def is_busy(self):
        self.lock.acquire()
        busy = self.isbusy
        self.lock.release()
        return busy

    # ------------------------------------------------------------------------------
    #           set_not_busy
    # ------------------------------------------------------------------------------

    def set_not_busy(self):

        self.lock.acquire()
        self.isbusy = False
        self.lock.release()

    # ------------------------------------------------------------------------------
    #           parse_list_of_directories
    # ------------------------------------------------------------------------------

    def parse_list_of_directories(self, htmlcontent: str):

        try:
            bytes = htmlcontent.encode()
        except:
            bytes = htmlcontent

        dirlisting, filelisting = self.decode_filelisting(bytes)
        self.threadrequest = None
        self.set_not_busy()
        self.callback(dirlisting)

    # ------------------------------------------------------------------------------
    #           parse_list_of_files
    # ------------------------------------------------------------------------------

    def parse_list_of_files(self, htmlcontent):

        try:
            bytes = htmlcontent.encode()
        except:
            bytes = htmlcontent

        dirlisting, filelisting = self.decode_filelisting(bytes)
        self.threadrequest = None
        self.set_not_busy()
        self.callback(self.dirname, filelisting)

    # ------------------------------------------------------------------------------
    #           fetch_list_of_directories
    # ------------------------------------------------------------------------------

    def fetch_list_of_directories(self, callback: Callable):
        """
        Uses an HTML request to fetch a list of directories.
        """

        ok = not self.is_busy()
        if (ok):
            self.wait_for_not_busy()
            self.callback = callback
            try:
                self.threadrequest = ThreadedHTMLRequest(self.globalobject.gui_app, self.html_config)
                self.threadrequest.dummydata = r'Index of _sdcard0_images_.htm'
                url = 'http://' + self.html_config["show_ip"] + '/sdcard0/images/'
                self.threadrequest.start_htmlfetch(url, self.parse_list_of_directories, timeout=self.html_config[
                    "timeout"])
            except:
                #           print('Directory listing exception thrown, releasing lock')
                self.set_not_busy()
        return ok

    # ------------------------------------------------------------------------------
    #           save_downloadfile_todisk
    # ------------------------------------------------------------------------------

    def save_downloadfile_todisk(self, htmlcontent):

        outputfilename = self.downloadparams[0]
        treeview_dirname = self.downloadparams[1]
        utcstr = self.downloadparams[2]
        item = self.downloadparams[3]

        folder, tail = os.path.split(outputfilename)
        if (not os.path.exists(folder)):
            os.makedirs(folder)
        f = open(outputfilename, 'wb')
        f.write(htmlcontent)
        f.close()
        self.downloadparams = None
        self.threadrequest = None
        self.set_not_busy()
        self.callback(outputfilename, treeview_dirname, utcstr, item)

    # ------------------------------------------------------------------------------
    #           fetch_image
    # ------------------------------------------------------------------------------

    def fetch_image(self, filename: str, treeview_dirname: str, utcstr: str, item: QtWidgets.QTreeWidgetItem, callback: Callable):

        ok = not self.is_busy()
        if (ok):

            showdir, showname = filename.split('/')
            outputfilename = self.output_folder + os.sep + "images" + os.sep + showdir + os.sep + showname
            self.downloadparams = [outputfilename, treeview_dirname, utcstr, item]
            self.callback = callback
            if os.path.exists(outputfilename):
                callback(outputfilename, treeview_dirname, utcstr, item)
            else:
                self.wait_for_not_busy()
                try:
                    self.threadrequest = ThreadedHTMLRequest(self.globalobject.gui_app, self.html_config, isbinaryfile=True)
                    self.threadrequest.dummydata = r'1491425056.194740-400ms-25C_WHITE_ARM_A.raw'
                    url = 'http://' + self.html_config["show_ip"] + '/sdcard0/images/' + filename
                    self.threadrequest.start_htmlfetch(url, self.save_downloadfile_todisk, timeout=self.html_config[
                        "timeout"])
                except:
                    self.set_not_busy()
        return ok

    # ------------------------------------------------------------------------------
    #           fetch_list_of_directories
    # ------------------------------------------------------------------------------

    def fetch_list_of_files(self, directoryname: str, callback: Callable):
        """
        Uses an HTML request to fetch a list of directories.
        """
        #        print('file listing, acquiring lock')
        ok = not self.is_busy()
        if (ok):
            self.wait_for_not_busy()
            self.callback = callback
            self.dirname = directoryname
            #        print('file listing, lock acquired')
            self.current_listing = None  # erase any listing we have from previous calls
            try:
                self.threadrequest = ThreadedHTMLRequest(self.globalobject.gui_app, self.html_config)
                self.threadrequest.dummydata = r'Index of _sdcard0_images_2017-04-25_19-26_.htm'
                url = 'http://' + self.html_config["show_ip"] + '/sdcard0/images/' + directoryname + '/'
                self.threadrequest.start_htmlfetch(url, self.parse_list_of_files, timeout=self.html_config["timeout"])
            except:
                #            print('File listing exception thrown, releasing lock')
                self.set_not_busy()
        return ok
