from typing import Dict, Any, List, Tuple, NamedTuple
from PyQt5 import QtCore, QtGui, QtWidgets
from abc import ABC, abstractmethod
import os.path
from .application import ArgolGSE_QtUI_Application
from .logger import LOG_Listener
from .queued_ftp import Argol_Queued_FTP, queued_ftp_download_entry
from .yaml_configuration import Yaml_Configuration

# THESE CANNOT BE GLOBAL AS THEY ATTEMPT TO LOAD ICONS BEFORE THE ICONS ARE LOADED INTO THE SYSTEM AND YOU END UP WITH NO ICONS ANYWHERE
# g_icon_fullfolder                 = QtGui.QIcon(":/icons/resources/icons/folder.ico")
# g_icon_halffullfolder             = QtGui.QIcon(":/icons/resources/icons/halffullfolder.ico")
# g_icon_emptyfolder                = QtGui.QIcon(":/icons/resources/icons/emptyfolder.ico")
# g_icon_fullfile                   = QtGui.QIcon(":/icons/resources/icons/fullfile.ico")
# g_icon_emptyfile                  = QtGui.QIcon(":/icons/resources/icons/emptyfile.ico")

Treeview_Itemdata = NamedTuple('Treeview_Itemdata', [('utcstr', str), ('directory', str),
                                                     ('displayelements', List[str]), ('displaynames', List[str]),
                                                     ('firmware_fullname', str), ('local_fullname', str)])


# ------------------------------------------------------------------------------
#           Treeview_Entry_Generator_Base
# ------------------------------------------------------------------------------

class Treeview_Entry_Generator_Base(ABC):
    """
    An actual implementation of the treeview must provide a class derived from this one that
    converts a text string from the instrument firmware into a Treeview_Itemdata
    """

    @abstractmethod
    def set_treeview(self, treeview: 'Treeview_Gui'):
        """
        Register the 'Treeview_Gui' display object with the generator class. This is often used to  add local files to the display
        """
        pass

    @abstractmethod
    def make_filename_entry(self, firmware_filename: str) -> Treeview_Itemdata:
        """
        Make a filename entry that can be displayed in the tree viw parent object. This consists of the information displayed in each column
        """
        pass

    @abstractmethod
    def get_header_labels(self) -> Tuple[List[str], List[int]]:
        """
        Get the labels and widths of the columns used to display each file
        """
        pass

    @abstractmethod
    def invoke_action_on_select(self, treeview_entry: Treeview_Itemdata):
        pass


# ------------------------------------------------------------------------------
#           Class _TreeView_ImageEntry
# ------------------------------------------------------------------------------

class _TreeView_ImageEntry:
    """
    Class used to store the python state of each item in the treeview display. The treeview is used to display info about
    files stored on the firmware
    """

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, itemvalue: Treeview_Itemdata, parent: 'Treeview_GUI'):
        """
        Creates the python object used to represent an item in the treeview
        """
        self.parent: 'Treeview_GUI' = parent
        self.value: Treeview_Itemdata = itemvalue  # Type: SHOWGSE_FileListingEntry, details about the image received from SHOW instrument
        self.is_local: bool = False  # Type: bool, True if the file is stored locally

    # ------------------------------------------------------------------------------
    #           status_icon
    # ------------------------------------------------------------------------------

    def status_icon(self, check_status: bool = False) -> QtGui.QIcon:

        if (check_status):
            self.is_local = os.path.exists(self.value.local_fullname)  # check the local status and update
        if self.is_local:
            icon = self.parent.g_icon_fullfile
        else:
            icon = self.parent.g_icon_emptyfile
        return icon


# ------------------------------------------------------------------------------
#           Class _TreeView_DirectoryEntry
# ------------------------------------------------------------------------------

class _TreeView_DirectoryEntry:

    # ------------------------------------------------------------------------------
    #               __init__
    # ------------------------------------------------------------------------------

    def __init__(self, treeview_dirname: str, parent: 'Treeview_GUI'):

        self.parent: 'Treeview_GUI' = parent
        self.dirname: str = treeview_dirname
        self.list_of_files: Dict[
            str, _TreeView_ImageEntry] = {}  # [str, ArgolGSE_FileListing_TreeView_ImageEntry], the string is the string displayed in the tree view
        self.load_status: int = 2

    # ------------------------------------------------------------------------------
    #           status_icon
    # ------------------------------------------------------------------------------

    def status_icon(self):

        if self.load_status == 2:
            icon = self.parent.g_icon_fullfolder
        elif self.load_status == 1:
            icon = self.parent.g_icon_halffullfolder
        else:
            icon = self.g_icon_emptyfolder
        return icon

    # ------------------------------------------------------------------------------
    #           find_or_create_entry
    # ------------------------------------------------------------------------------

    def find_or_create_file_entry(self, value: Treeview_Itemdata) -> _TreeView_ImageEntry:

        entry = self.list_of_files.get(value.utcstr)
        if (entry is None):
            self.list_of_files[value.utcstr] = _TreeView_ImageEntry(value, self.parent)
            entry = self.list_of_files[value.utcstr]
        return entry

    # ------------------------------------------------------------------------------
    #           find_or_create_entry
    # ------------------------------------------------------------------------------

    def find_file_entry(self, utcstr: str) -> _TreeView_ImageEntry:

        entry = self.list_of_files.get(utcstr)
        return entry

    # ------------------------------------------------------------------------------
    #           add_file
    # ------------------------------------------------------------------------------

    def add_file(self, value: Treeview_Itemdata,
                 treeview: QtWidgets.QTreeWidget) -> _TreeView_ImageEntry:

        treeitems = treeview.findItems(value.directory, QtCore.Qt.MatchFixedString)  # find this directory entry in our the GUI treeview widget
        directoryitem = treeitems[0]  # select the first one. There should be only one
        entry = self.find_or_create_file_entry(value)  # Find the file entry in the python object
        treeitems = treeview.findItems(value.utcstr, QtCore.Qt.MatchFixedString | QtCore.Qt.MatchRecursive)  # and make sure we cannot find this file object in the current tree view
        if len(treeitems) == 0:  # If we cannot find it
            fileitem = QtWidgets.QTreeWidgetItem(directoryitem, value.displayelements, QtWidgets.QTreeWidgetItem.Type)  # then create the tree-view entry for this file object
        else:
            fileitem = treeitems[0]
        fileitem.setIcon(0, entry.status_icon(check_status=True))
        directoryitem.setIcon(0, self.status_icon())  # and set the status of the dirctory icon.
        return entry

    # ------------------------------------------------------------------------------
    #           add_file
    # ------------------------------------------------------------------------------

    def select_file(self, value: Treeview_Itemdata,
                    treeview: QtWidgets.QTreeWidget) -> Tuple[_TreeView_ImageEntry, QtWidgets.QTreeWidgetItem]:

        fileitem = None
        entry = self.find_file_entry(value.utcstr)  # Find the file entry in the internal directory object
        if (entry is not None):
            treeitems = treeview.findItems(value.utcstr, QtCore.Qt.MatchFixedString | QtCore.Qt.MatchRecursive)  # Find this file object in the current tree view
            if len(treeitems) > 0:  # If we fount it
                fileitem = treeitems[0]  # get the first item found
                fileitem.setIcon(0, entry.status_icon(check_status=True))  # and set its icon status.
            else:
                print("_TreeView_DirectoryEntry::select_file, could not find treeview entry in widget for ", value.utcstr, '. Thats not good.')
        else:
            print("_TreeView_DirectoryEntry::select_file, could not find file entry in internal directory list for ", value.utcstr, '. Thats not good.')
        return entry, fileitem

    # ------------------------------------------------------------------------------
    #           update_subfolder_file_status
    # ------------------------------------------------------------------------------

    def update_subfolder_file_status(self, treeview: QtWidgets.QTreeWidget):

        for utcstr, entry in self.list_of_files.items():  # for each file in this directory
            fileitem = treeview.findItems(entry.value.utcstr, QtCore.Qt.MatchFixedString | QtCore.Qt.MatchRecursive)  # find this item in the tree
            if len(fileitem) != 0:  # and if we found it
                fileitem[0].setIcon(0, entry.status_icon(check_status=True))  # then set its icon status


# ------------------------------------------------------------------------------
#           ArgolGSE_FileListingGui
# ------------------------------------------------------------------------------

class Treeview_GUI:
    """
    A treeview used to display files generated by the argol firmware applications. The basic idea is that the firmware 
    sends a special log message to the ground such as,
    `..\cats_test_data\20180314\20180314\cats-ccd\20180314-2214\cats-ccd_heightscan_20180314-221409.448000_v001.fits`.
    This is a notification that the firmware has just created this file on disk. The Treeview GUI system is then used to
    provide visual feedback to the GSE user. The interface provides options to automatically download files as the
    notifications are received and to manually download images upon (clicked) request. The treeview GUI has to maintain
    two lists, (1) the QT gui tree view and (2) the python representtation of that list.

    New entries are added to the treeview in two ways: (1) through an explicit call to add_new_entry or (2) from special notfications
    received by the logger system. Special notifications in the logger system are special messages sent by the firmware
    over the logging system indicating that a new file has been generated on the firmware. The special message is  decoded by the
    specific instance of Treeview_Entry_Generator_Base passed into the constructor.
    """

    def __init__(self, qtapp: ArgolGSE_QtUI_Application,
                 treeview_widget_name: str,
                 itemdata_generator: Treeview_Entry_Generator_Base,
                 logger: LOG_Listener,
                 speciallogstr: str,
                 ftpdownloader: Argol_Queued_FTP):

        self.downloadmode: str = 'MANUAL'  # How should we implement FTP download
        self.itemdata_generator: Treeview_Entry_Generator_Base = itemdata_generator
        self.directories: Dict[str, _TreeView_DirectoryEntry] = {}
        self.treeview: QtWidgets.QTreeWidget = getattr(qtapp.ui, treeview_widget_name)
        self.lightgraybrush: QtGui.QBrush = QtGui.QBrush(QtCore.Qt.lightGray)
        self.darkgraybrush: QtGui.QBrush = QtGui.QBrush(QtCore.Qt.darkGray)
        self.g_icon_fullfolder: QtGui.QIcon = QtGui.QIcon(":/icons/icons/folder.ico")
        self.g_icon_halffullfolder: QtGui.QIcon = QtGui.QIcon(":/icons/icons/halffullfolder.ico")
        self.g_icon_emptyfolder: QtGui.QIcon = QtGui.QIcon(":/icons/icons/emptyfolder.ico")
        self.g_icon_fullfile: QtGui.QIcon = QtGui.QIcon(":/icons/icons/fullfile.ico")
        self.g_icon_emptyfile: QtGui.QIcon = QtGui.QIcon(":/icons/icons/emptyfile.ico")
        self.ftpdownload: Argol_Queued_FTP = ftpdownloader

        self._configure_tree_view()
        self.treeview.itemPressed.connect(self.item_pressed)
        self.treeview.itemExpanded.connect(self.itemExpanded)
        itemdata_generator.set_treeview(self)

        self._set_download_mode_from_configuration(speciallogstr)  # See if we are Auto, Manual of Off for FTP download
        if (speciallogstr is not None):
            logger.register_special_callback(speciallogstr, self._insert_archive_into_treeview)

        # ------------------------------------------------------------------------------
        #           _load_special_types
        # ------------------------------------------------------------------------------

    def _set_download_mode_from_configuration(self, speciallogstr: str):

        self.downloadmode = 'MANUAL'  # How should we implement FTP download
        if (speciallogstr is not None):  # Do we have a special log str (eg ARCHIVE or SAMPELD)
            specialtypes = Yaml_Configuration('network.yaml').configuration['FTP'][
                'specialtypes']  # Yup, so see if we have a download defaul
            if (specialtypes is not None):  # If we have a configuration entry
                mode = specialtypes[speciallogstr]  # Then get the entry
                if (mode is None):
                    mode = self.downloadmode  # and replace with default
                self.downloadmode = mode.upper()  # Get the upper case version only
        ok = (self.downloadmode == "MANUAL") or (self.downloadmode == "AUTO") or (self.downloadmode == "OFF")
        if (not ok):
            print("Treeview_GUI:: _set_download_mode_from_configuration, The download mode [", self.downloadmode, "] is not supported. Replacing with MANUAL")
            self.downloadmode = "MANUAL"

    # ------------------------------------------------------------------------------
    #           insert_archive_into_treeview
    # ------------------------------------------------------------------------------

    def _insert_archive_into_treeview(self, leader: str, remote_filename: str):
        self.add_new_entry(remote_filename)

    # ------------------------------------------------------------------------------
    #           configure_tree_view
    # ------------------------------------------------------------------------------

    def _configure_tree_view(self):

        labels, width = self.itemdata_generator.get_header_labels()
        self.treeview.setHeaderLabels(labels)
        for i in range(len(width)):
            self.treeview.setColumnWidth(i, width[i])

    # ------------------------------------------------------------------------------
    #           directory_entry
    # ------------------------------------------------------------------------------

    def _find_or_create_python_directory_entry(self, dirname: str):

        entry = None
        if (dirname is not None):
            entry = self.directories.get(dirname)
            if entry is None:
                self.directories[dirname] = _TreeView_DirectoryEntry(dirname, self)
                entry = self.directories[dirname]
        return entry

    # ------------------------------------------------------------------------------
    #           fetch_or_create_widget_directory_item
    # ------------------------------------------------------------------------------

    def _fetch_or_create_widget_directory_item(self, entry: _TreeView_DirectoryEntry) -> QtWidgets.QTreeWidgetItem:

        treeitems = self.treeview.findItems(entry.dirname, QtCore.Qt.MatchFixedString)
        if (len(treeitems) > 0):
            item = treeitems[0]
        else:
            item = QtWidgets.QTreeWidgetItem([entry.dirname, ''], QtWidgets.QTreeWidgetItem.Type)
            item.setChildIndicatorPolicy(QtWidgets.QTreeWidgetItem.DontShowIndicatorWhenChildless)
            item.setIcon(0, entry.status_icon())
            self.treeview.addTopLevelItem(item)
        return item

    # ------------------------------------------------------------------------------
    #           download_complete_callback
    # ------------------------------------------------------------------------------

    def download_complete_callback(self, ftpentry: queued_ftp_download_entry):
        """
        Callback from the FTP download whenthe download has completed. This will
        only be called after the download has completed and will never be called if the
        FTP transfer never properly connects etc.
        """
        fullname = ftpentry.remote_filename
        value: Treeview_Itemdata = self.itemdata_generator.make_filename_entry(fullname)
        direntry: _TreeView_DirectoryEntry = self._find_or_create_python_directory_entry(value.directory)
        print('download_complete_callback has been called')
        direntry.select_file(value, self.treeview)

    # ------------------------------------------------------------------------------
    #           add_new_entry
    # ------------------------------------------------------------------------------

    def add_new_entry(self, remote_filename: str):

        value = self.itemdata_generator.make_filename_entry(remote_filename)
        direntry = self._find_or_create_python_directory_entry(value.directory)
        if (direntry is not None):
            self._fetch_or_create_widget_directory_item(direntry)
            entry = direntry.add_file(value, self.treeview)
            if (not entry.is_local):
                if (self.downloadmode == "AUTO") and (
                        self.ftpdownload is not None):  # If we are using automatic download
                    ftpentry = queued_ftp_download_entry(remote_filename=entry.value.firmware_fullname,
                                                         local_fullname=entry.value.local_fullname,
                                                         download_complete_callback=self.download_complete_callback)
                    self.ftpdownload.queue_file(ftpentry)
                # Then initiate automatic download

    # ------------------------------------------------------------------------------
    #           item_pressed
    # ------------------------------------------------------------------------------

    @QtCore.pyqtSlot(QtWidgets.QTreeWidgetItem, int)
    def item_pressed(self, item: QtWidgets.QTreeWidgetItem, column: int):

        text = item.text(0)
        parent = item.parent()
        if (parent is None):  # Is this a top level item/directory
            if item.isExpanded():
                direntry = self.directories.get(text)
                if (direntry is not None):
                    direntry.update_subfolder_file_status(self.treeview)
        else:
            # print('File Item pressed in Tree view, text = [', text,']')
            dirname = parent.text(0)
            direntry = self.directories.get(dirname)
            if direntry is not None:
                file_entry = direntry.find_file_entry(text)
                if (file_entry is not None):
                    if (file_entry.is_local):
                        self.itemdata_generator.invoke_action_on_select(file_entry.value)
                    else:
                        if (self.ftpdownload is not None) and (
                                (self.downloadmode == "AUTO") or (self.downloadmode == "MANUAL")):
                            ftpentry = queued_ftp_download_entry(remote_filename=file_entry.value.firmware_fullname,
                                                                 local_fullname=file_entry.value.local_fullname,
                                                                 download_complete_callback=self.download_complete_callback)

                            self.ftpdownload.queue_file(ftpentry)

    # ------------------------------------------------------------------------------
    #           itemExpanded
    # ------------------------------------------------------------------------------

    @QtCore.pyqtSlot(QtWidgets.QTreeWidgetItem)
    def itemExpanded(self, item: QtWidgets.QTreeWidgetItem):
        text = item.text(0)
        parent = item.parent()
        if (parent is None):  # Is this a top level item/directory
            direntry = self.directories.get(text)
            if (direntry is not None):
                direntry.update_subfolder_file_status(self.treeview)
