from typing import Dict, Any, Tuple, List, Union, Callable
from enum import Enum
import datetime
from PyQt5 import QtCore, QtGui, QtWidgets
from ..application import ArgolGSE_QtUI_Application
from ..cmd import CMD_Manager, CommandStatusLight, Command_History_Entry
from ..yaml_configuration import Yaml_Configuration
from ..widgets.sensor_variable_widgets import SensorVariableUI_String
from ..widgets.tab_widget_manager import Tab_Widget_Manager


# ------------------------------------------------------------------------------
#           SHOWGSE_CommandGuiEntry_ComboBox
# ------------------------------------------------------------------------------

class Command_Gui(QtCore.QObject):
    """
    An object that couples a QT QComboBox to a SHOW command state expressed as a string.
    The Command state can be either a fixed list of options (e.g. On and Off) or may have
    and editable values. The SHOW command state should be an instance of SHOWGSE_CommandStateVariableSelect
    """

    def __init__(self, qtapp: ArgolGSE_QtUI_Application, tab_widget_manager: Tab_Widget_Manager, config_folder=None):

        super().__init__()

        self.cmd_manager: CMD_Manager = CMD_Manager(qtapp, config_folder=config_folder)
        self.commands_config: Yaml_Configuration = Yaml_Configuration('commands.yaml', basedir=config_folder)
        self.sumbmodule_widget: QtWidgets.QComboBox = qtapp.ui.CMD_COMBO_SUBMODULE
        self.command_widget: QtWidgets.QComboBox = qtapp.ui.CMD_COMBO_COMMAND
        self.parameter_widget: QtWidgets.QComboBox = qtapp.ui.CMD_LINEEDIT_PARAMETERS
        # self.command_status_widget : QtWidgets.QLabel      = qtapp.ui.CMD_PENDING_ICON
        self.send_button_widget: QtWidgets.QPushButton = qtapp.ui.CMD_BUTTON_SEND
        self.cmd_history_widget: QtWidgets.QTableWidget = qtapp.ui.CMD_TEXTEDIT_HISTORY
        self.cmd_send_udp_port: SensorVariableUI_String = SensorVariableUI_String('CMD_SEND_UDP_PORT', 'str', qtapp, tab_widget_manager, None, options={'tabicon': "voltages_currents:HK"})
        self.cmd_echo_udp_port: SensorVariableUI_String = SensorVariableUI_String('CMD_ECHO_UDP_PORT', 'str', qtapp, tab_widget_manager, None, options={'tabicon': "voltages_currents:HK"})
        self.gse_ip: QtWidgets.QLabel = getattr(qtapp.ui, 'GSE_IP_ADDRESS')

        self.trafficlights: Dict[CommandStatusLight, QtGui.QIcon] = None

        self.g_sendlight = [QtGui.QPixmap(":/icons/icons/statusmed_green.ico"),  # TrafficLight.Green
                            QtGui.QPixmap(":/icons/icons/statusmed_orange.ico"),  # TrafficLight.Orange
                            QtGui.QIcon(":/icons/icons/statusmed_green.ico"),  # TrafficLight.Green
                            QtGui.QIcon(":/icons/icons/statusmed_orange.ico")  # TrafficLight.Orange
                            ]

        self.cmd_manager.add_callback(self.command_received)
        self.send_button_widget.clicked.connect(self.send_button_clicked)
        self.sumbmodule_widget.currentTextChanged.connect(self._select_submodule_textchanged)
        self.command_widget.currentTextChanged.connect(self._select_parameter_textchanged)

        self._initialize_combo_boxes()

        self.cmd_send_udp_port.set_value(str(self.cmd_manager.cmd_send.udp.udp_port_inst))
        self.cmd_echo_udp_port.set_value(str(self.cmd_manager.cmd_echo.udp.udp_port_inst))
        if (self.gse_ip is not None):
            self.gse_ip.setText(str(self.cmd_manager.cmd_echo.udp.udp_ipaddress))

    # ------------------------------------------------------------------------------
    #           command_received
    # ------------------------------------------------------------------------------

    def command_received(self, cmdindex: int,
                         cmdstr: str,
                         recvack: bool,
                         cmdack: bool,
                         goodstatus: bool):

        entry = None
        # self.command_status_widget.setPixmap(self.g_sendlight[0])
        self.send_button_widget.setIcon(self.g_sendlight[2])
        # print("Command received: index = ", cmdindex, ": recv ack = ", recvack, " cmdack = ", cmdack, 'status = ', goodstatus, " str = ", cmdstr)
        try:
            entry = self.cmd_manager.command_history[cmdindex]
            if (recvack):
                if (goodstatus):
                    entry.receive_status = CommandStatusLight.Good
                else:
                    entry.receive_status = CommandStatusLight.Bad
                    entry.command_status = CommandStatusLight.Bad
                entry.reply = cmdstr
            if (cmdack):
                if (goodstatus):
                    entry.command_status = CommandStatusLight.Good
                else:
                    entry.command_status = CommandStatusLight.Bad
                entry.reply = cmdstr
        except:
            if (cmdstr != "HK PRINTHEADER"):
                print("GSE Warning: I have no history of the command index ", cmdindex, ". Thats odd. It can happen when a custom command is sent.  recv ack = ", recvack, " cmdack = ", cmdack, 'status = ', goodstatus, " str = ", cmdstr)
            entry = None

        if (entry is not None):
            self.update_new_history_entry(entry)

    # ------------------------------------------------------------------------------
    #           add_new_history_entry
    # ------------------------------------------------------------------------------

    def update_new_history_entry(self, entry: Command_History_Entry):

        nrows = entry.tableview_row
        if (nrows >= 0):
            item = []
            item.append(self.cmd_history_widget.item(nrows, 0))
            item.append(self.cmd_history_widget.item(nrows, 1))
            item.append(self.cmd_history_widget.item(nrows, 2))
            item.append(self.cmd_history_widget.item(nrows, 3))
            item.append(self.cmd_history_widget.item(nrows, 4))

            # item[0].setText(entry.utc.strftime("%H:%M:%S"))
            if (entry.receive_status == CommandStatusLight.Good):
                item[1].setText('RECV OK')
            elif (entry.receive_status == CommandStatusLight.Bad):
                item[1].setText('RECV BAD')
            else:
                item[1].setText('RECV WAIT')

            if (entry.command_status == CommandStatusLight.Good):
                item[2].setText('CMD OK')
            elif (entry.command_status == CommandStatusLight.Bad):
                item[2].setText('CMD BAD')
            else:
                item[2].setText('CMD WAIT')

            self.cmd_history_widget.item(nrows, 1).setIcon(self.icon_from_status(entry.receive_status))
            self.cmd_history_widget.item(nrows, 2).setIcon(self.icon_from_status(entry.command_status))

            item[4].setText(entry.reply)
            self.cmd_history_widget.scrollToBottom()

    # ------------------------------------------------------------------------------
    #           add_new_history_entry
    # ------------------------------------------------------------------------------

    def add_new_history_entry(self, entry: Command_History_Entry):

        nrows = 0
        if (entry.tableview_row < 0):
            nrows = self.cmd_history_widget.rowCount()
        entry.tableview_row = nrows
        self.cmd_history_widget.insertRow(nrows)
        self.cmd_history_widget.setItem(nrows, 0, QtWidgets.QTableWidgetItem(entry.utc.strftime("%Y-%m-%d %H:%M:%S.%f")))
        self.cmd_history_widget.setItem(nrows, 1, QtWidgets.QTableWidgetItem('RECV Wait'))
        self.cmd_history_widget.setItem(nrows, 2, QtWidgets.QTableWidgetItem('CMD Wait'))
        self.cmd_history_widget.setItem(nrows, 3, QtWidgets.QTableWidgetItem(entry.command))
        self.cmd_history_widget.setItem(nrows, 4, QtWidgets.QTableWidgetItem())

        self.cmd_history_widget.item(nrows, 1).setIcon(self.icon_from_status(entry.receive_status))
        self.cmd_history_widget.item(nrows, 2).setIcon(self.icon_from_status(entry.command_status))

        self.cmd_history_widget.resizeColumnToContents(0)
        self.cmd_history_widget.resizeColumnToContents(1)
        self.cmd_history_widget.resizeColumnToContents(2)
        self.cmd_history_widget.resizeColumnToContents(3)
        self.cmd_history_widget.scrollToBottom()

    # ------------------------------------------------------------------------------
    #           icon_from_status
    # ------------------------------------------------------------------------------

    def icon_from_status(self, status: CommandStatusLight) -> QtGui.QIcon:

        if (self.trafficlights is None):
            self.trafficlights = {CommandStatusLight.Good: QtGui.QIcon(":/icons/icons/statusmed_green.png"),
                                  CommandStatusLight.Waiting: QtGui.QIcon(":/icons/icons/statusmed_orange.png"),
                                  CommandStatusLight.Bad: QtGui.QIcon(":/icons/icons/statusmed_red.png"),
                                  }

        icon = self.trafficlights[status]
        return icon

    # ------------------------------------------------------------------------------
    #           send_button_clicked
    # ------------------------------------------------------------------------------

    @QtCore.pyqtSlot(bool)
    def send_button_clicked(self, clicked: bool):

        submodule = self.sumbmodule_widget.currentText()
        command = self.command_widget.currentText()
        params = self.parameter_widget.currentText()
        cmd = submodule + ' ' + command + ' ' + params
        cmd.strip()
        id = self.cmd_manager.send_command(cmd)
        # self.command_status_widget.setPixmap( self.g_sendlight[1])
        self.send_button_widget.setIcon(self.g_sendlight[3])
        self.add_new_history_entry(self.cmd_manager.command_history[id])

    # ------------------------------------------------------------------------------
    #           _configure_combo_boxes
    # ------------------------------------------------------------------------------

    def _initialize_combo_boxes(self):

        submodulekeys = self.commands_config.configuration.keys()

        for submodule in submodulekeys:
            self.sumbmodule_widget.addItem(submodule)

    # ------------------------------------------------------------------------------
    #           _select_submodule_textchanged
    # ------------------------------------------------------------------------------
    @QtCore.pyqtSlot(str)
    def _select_submodule_textchanged(self, submodule: str):

        if submodule in self.commands_config.configuration.keys():
            commandlist: List[str] = self.commands_config.configuration[submodule]
            self.command_widget.clear()
            self.parameter_widget.clear()
            for command in commandlist:
                tokens = command.split('[')
                ntokens = len(tokens)
                options = []
                if (ntokens > 1):
                    for i in range(ntokens - 1):
                        o = tokens[i + 1].strip(' ][')
                        options.append(o)
                self.command_widget.addItem(tokens[0].strip(), options)

    # ------------------------------------------------------------------------------
    #           _select_submodule_textchanged
    # ------------------------------------------------------------------------------
    @QtCore.pyqtSlot(str)
    def _select_parameter_textchanged(self, parameter: str):

        index = self.command_widget.findText(parameter)
        if (index > -1):
            optionlist = self.command_widget.currentData()
            self.parameter_widget.clear()
            for option in optionlist:
                self.parameter_widget.addItem(option)
