import math
import numpy as np
import logging
from typing import Dict, Any, Tuple, List, Union, Callable
from abc import ABC, abstractmethod
from enum import IntEnum
from PyQt5 import QtCore, QtGui, QtWidgets
from ..application import ArgolGSE_QtUI_Application
from .tab_widget_manager import Tab_Widget_Manager
from ..cmd import CMD_Manager
from .trafficlight import TrafficLight


# #-----------------------------------------------------------------------------------------------------------------------
# #           TrafficLight
# #   A simple enumeration that maps a traffic light color to a display icon.
# #-----------------------------------------------------------------------------------------------------------------------
#
# class TrafficLight(IntEnum):
#     """
#     A small enumeration to hold  the state of the GUI of the traffic light statuses.
#     Some code may rely upon the ordering so dont change the numerical ordering. The code relies upon the fact that more severe errors
#     have a bigger number than less severe errors.
#     """
#
#     Undefined    = 0        #
#     Green        = 1        #
#     Green_Orange = 2        #
#     Orange       = 3        #
#     Red          = 4        #
#     Blue         = 5        #

# ------------------------------------------------------------------------------
#           _statusleveltotrafficlight
# ------------------------------------------------------------------------------

def _statusleveltotrafficlight(status: str) -> TrafficLight:
    value = status.upper()
    light = TrafficLight.Undefined
    if (value == "GREEN"):
        light = TrafficLight.Green
    elif (value == "RED"):
        light = TrafficLight.Red
    elif (value == "BLUE"):
        light = TrafficLight.Blue
    elif (value == "UNDEFINED"):
        light = TrafficLight.Undefined
    elif (value == "ORANGE"):
        light = TrafficLight.Orange
    elif (value == "GREEN_ORANGE"):
        light = TrafficLight.Green_Orange
    return light


# -----------------------------------------------------------------------------------------------------------------------
#           class SensorVariableUI_Base
# -----------------------------------------------------------------------------------------------------------------------

class SensorVariableUI_Base(ABC):
    """
    The base class for the Argol SensorVariable User interface objects. The *SensorVariableUI* interface consists of
    a QLabel widget that displays the variables value as a string plus two optional GUI elements, a traffic light icon and a push-button.
    The traffic light icon is implemented as the **pixmap** of a QLabel to display the ``FAIL/MARGINAL/PASS`` status of the variable using a
    traffic light icon paradigm. The third widget is an optional QPushButton widget which can be programmed in derived classes
    to send commands with a :class:`~.CMD_Manager`. It is typically used to toggle relays, power switches, power off computer etc.

    The whole system is driven by a yaml file which contains a section for each sensor variable. The name of each section
    is usually set up so it is both the name of the python variable in an associated dictionary and the name of the primary
    QLabel widget in the QT Designer user interface.  The name of the traffic light QLabel:pixmap user interface element
    is set by appending ``_ICON`` to the variable name.  The code will ignore the traffic light sections of code if the
    interface element does not exist in the compiled ``ui`` file. Similarly, the name of the push button user interface element
    is set by appending ``_BUTTON`` to the variable name.  The code will ignore the push-button sections of code if the
    interface element does not exist in the compiled ``ui`` file. The code allows the yaml file to override the default
    naming convention using the *hk_name* and *ui_names** attributes.

    The class also manages displaying a traffic light icon in the associated tabbed dialog page. The concept is that this
    icon is always visible to the operator even if the tabbed dialog contents are not. The tabbed dialog icon will always
    display the most severe condition from all of the contributing sensor variables. The *tabicon* attribute in the yaml file
    controlsthis option.

    Parameters
    -----------
    sensor_name: str
        The (default) name of this sensor variable. The name is usually taken from a yaml section

    type: str
        The type of SensorVariableUI widget we are creating. It must be a class derived from ``SensorVariableUI_Base``

    qtapp: ArgolGSE_QtUI_Application
        The QT application

    tabiconmanager: Tab_Widget_Manager
        The instance of :class:`~.Tab_Widget_Manager` that is managing the tabbed dialogs in this gui application

    cmd_mgr: CMD_Manager,
        The instance of :class:`~.CMD_Manager` that is managing commands in this gui application.

    options: Dict[str, Any]
        The dictionary of attributes. This is usally taken from the settings within the associated section of the yaml file.
    """

    g_trafficlights = None

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, sensor_name: str,
                 type: str,
                 qtapp: ArgolGSE_QtUI_Application,
                 tabiconmanager: Tab_Widget_Manager,
                 cmd_mgr: CMD_Manager,
                 options: Dict[str, Any]):
        """
        Constructs the GUI object associated with this object. The `qtapp` is the QT main application that contains the QT
        UI object. This object will attempt to connect the sensor variable object with between one and three Widget
        objects. The first widget is a mandatory QLabel (or similar) widget used to display the value of the sensor variable.
        The second widget is an optional QIcon used to display the FAIL/PASS status of the sensor variable using a
        traffic light icon paradigm. The third widget is an optional QButton widget which can be programmed in derived
        classes to repond to button presses, typically used to quickly turn instrument SYSCTL relays on and off.
        """
        tabiconname = options.get('tabicon')
        if (tabiconname is not None):
            tokens = tabiconname.split(':')
        else:
            tokens = [None, None]

        self._sensor_name: str = sensor_name                                                # The name of this sensor variable. It ypically matches the name of a Qlabel in the dialog box,
        self.type: str = type                                                               # The type of the widget. This normally comes from the housekeeping command HK PRINTHEADERS
        self.cmd_manager: CMD_Manager = cmd_mgr
        self.tabiconname: str = tokens[0]                                                   # The name of the tab whose icon we will control (may be None)
        self.tabiconsection: str = tokens[1]                                                # The section of the tab icon status that this variable is compared in
        self.tabiconmanager: Tab_Widget_Manager = tabiconmanager                            # The manager who controls the icon status of each tab in the GSE display
        self._widget_value: QtWidgets.QLabel = None                                         # The QT QWidget, typically a QLabel, that is used to display the value of the sensor variable (typically from house-keeping)
        self._widget_icon: QtWidgets.QLabel = None                                          # The optional widget used  to visually display the PASS/FAIL status of the sensor variable as a traffic light icon
        self._widget_button: QtWidgets.QPushButton = None                                   # An optional QT QButton control used to executes commands associated with this sensor variable. Used to quickly switch relays on and off
        self._status: TrafficLight = TrafficLight.Undefined
        self._text: str = '><'
        self.hk_name: str = options.get('hk_name')
        self._update_callback: List[Callable] = []

        ui_names = options.get('ui_names')
        if (self.hk_name is None):
            self.hk_name = sensor_name
        self._attach_widgets(qtapp, ui_names)

    # ------------------------------------------------------------------------------
    #           icon_fromstatus
    # ------------------------------------------------------------------------------

    def _icon_from_status(self, status: TrafficLight) -> QtGui.QPixmap:
        """
        Returns the QT Gui traffic light icon corresoponding to the requested state value.
        """

        if (self.g_trafficlights is None):
            self.g_trafficlights = {TrafficLight.Green: QtGui.QPixmap(":/icons/icons/trafficlight_green.ico"),                  # TrafficLight.Green
                                    TrafficLight.Green_Orange: QtGui.QPixmap(":/icons/icons/trafficlight_greenorange.ico"),     # TrafficLight.Green-Orange
                                    TrafficLight.Orange: QtGui.QPixmap(":/icons/icons/trafficlight_orange.ico"),                # TrafficLight.Orange
                                    TrafficLight.Red: QtGui.QPixmap(":/icons/icons/trafficlight_red.ico"),                      # TrafficLight.Red
                                    TrafficLight.Blue: QtGui.QPixmap(":/icons/icons/trafficlight_blue.ico"),                    # TrafficLight.Blue
                                    TrafficLight.Undefined: QtGui.QPixmap(":/icons/icons/nolights.ico")                         # TrafficLight.Undefined
                                    }
        icon = self.g_trafficlights[status]
        return icon

    # ------------------------------------------------------------------------------
    #           attach_widgets
    # ------------------------------------------------------------------------------

    def _attach_widgets(self, qtapp: ArgolGSE_QtUI_Application, qt_ui_names: Tuple[str, str, str]):

        widgetnames = dir(qtapp.ui)

        if qt_ui_names is None:                                                                                         # If the User has not passed in explicit widget names
            widgetname = self.sensor_name                                                                               # then set up the defaults
            iconname = widgetname + '_ICON'
            buttonname = widgetname + '_BUTTON'
        else:
            l = len(qt_ui_names)
            widgetname = qt_ui_names[0]                                                                                 # We must have at least one entry
            if (l > 1):
                iconname = qt_ui_names[1]
            else:
                iconname = None
            if (l > 2):
                buttonname = qt_ui_names[2]
            else:
                buttonname = None
        # print("widgetname = [", widgetname,"]")
        try:
            self._widget_value = getattr(qtapp.ui, widgetname)                                                          # Find the Label (like) widget in the QT UI file. This must exist or crassh
        except AttributeError as e:
            self._widget_value = None

        if (iconname is not None) and (iconname in widgetnames):                                                        # If we are using an icon and it exists in the list
            self._widget_icon = getattr(qtapp.ui, iconname)                                                             # then get the widget icon
            self._widget_icon.setPixmap(self._icon_from_status(TrafficLight.Undefined))                                 # and set the icon to undefined (all lights off)

        if (buttonname is not None) and (buttonname in widgetnames):                                                    # if we are using a button and it exists
            self._widget_button = getattr(qtapp.ui, buttonname)                                                         # then attach the button

    # ------------------------------------------------------------------------------
    #           connect_button
    # ------------------------------------------------------------------------------

    def connect_button(self, func: Callable):

        if (self._widget_button is not None):
            self._widget_button.clicked.connect(func)

    # ------------------------------------------------------------------------------
    #           status Property
    # ------------------------------------------------------------------------------

    @property
    def status(self) -> TrafficLight:
        """
        The internal traffic light status of the sensor variable and sets the displayed icon
        """
        return self._status

    @status.setter
    def status(self, newstatus: TrafficLight):

        if newstatus != self._status:
            self._status = newstatus
            if self._widget_icon is not None:
                self._widget_icon.setPixmap(self._icon_from_status(self._status))
        if (self.tabiconname is not None):
            self.tabiconmanager.update_tab_status(self.tabiconname, self.tabiconsection, newstatus)

    # ------------------------------------------------------------------------------
    #           Property text
    # ------------------------------------------------------------------------------

    @property
    def text(self) -> str:
        """
        The text property of the sensor variable widget cluster. This is the text displayed in the QLabel widget
        """
        return self._text

    @text.setter
    def text(self, text: str):
        self._text = text
        if self._widget_value is not None:
            self._widget_value.setText(text)
        self._notify_callbacks()

    # ------------------------------------------------------------------------------
    #           Property sensor_name
    # ------------------------------------------------------------------------------
    @property
    def sensor_name(self) -> str:
        """
        The sensor variable associated with this widget cluster. This is the name associated with the sensor variables sent by the instrument firmware
        """
        return self._sensor_name

    # ------------------------------------------------------------------------------
    #           set_value
    # ------------------------------------------------------------------------------

    @abstractmethod
    def set_value(self, textstring: str):
        """
        Set the new value of the widget from the incoming string
        """
        pass

    # ------------------------------------------------------------------------------
    #           add_update_callback
    # ------------------------------------------------------------------------------

    def add_update_callback(self, func: Callable):
        """
        Adds a callback function to the list of functions called after the value changes
        """
        self._update_callback.append(func)

    # ------------------------------------------------------------------------------
    #           notify_callbacks
    # ------------------------------------------------------------------------------

    def _notify_callbacks(self):
        for f in self._update_callback:
            f(self.text)


# ------------------------------------------------------------------------------
#           SensorVariableUI_String
# ------------------------------------------------------------------------------

class SensorVariableUI_String(SensorVariableUI_Base):
    """
    Displays text values in a single QLabel element. This component does not interact with any traffic light elements
    and always returns a ``Green`` status if requested.
    """

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, sensor_name: str,
                 type: str,
                 qtapp: ArgolGSE_QtUI_Application,
                 tabiconmanager: Tab_Widget_Manager,
                 cmd_mgr: CMD_Manager,
                 options: Dict[str, Any]):
        super().__init__(sensor_name, type, qtapp, tabiconmanager, cmd_mgr, options)

    # ------------------------------------------------------------------------------
    #           set_value
    # ------------------------------------------------------------------------------

    def set_value(self, newvalue: str):
        self.text = newvalue
        self.status = TrafficLight.Green


# ------------------------------------------------------------------------------
#           SensorVariable_BoundedFloat
# ------------------------------------------------------------------------------

class SensorVariableUI_BoundedFloat(SensorVariableUI_Base):
    """
    Displays floating point values that should be bounded and is intended to display variables like voltages
    and temperatures which should stay within well defined bounds. The class defines a green, orange and red  zones:
    green is well within comfortable operation, orange is still acceptable but is approaching it upper or lower limit
    boundary while red indicates the variable exceeds its acceptable operating boundaries.
    """

    def __init__(self, sensor_name: str,
                 type: str,
                 qtapp: ArgolGSE_QtUI_Application,
                 tabiconmanager: Tab_Widget_Manager,
                 cmd_mgr: CMD_Manager,
                 options: Dict[str, Any]):

        super().__init__(sensor_name, type, qtapp, tabiconmanager, cmd_mgr, options)

        self.number: float = math.nan  # The current numeric value.
        self.range: List[
            float, float, float, float] = options.get('range')  # The minimum values for orange and red zone boundaries.
        self.format: str = options.get('format')

        if (len(self.range) % 4) != 0:
            logging.warning('SensorVariableUI_BoundedFloat, Floating point widgets for %s must set the range configuration keyword to a multiple of 4 list elements ' % (sensor_name,))
        if (self.format is None):
            self.format = "%6.2f"

    # ------------------------------------------------------------------------------
    #           set_value(self, newvalue: str, utc: datetime.datetime):
    # ------------------------------------------------------------------------------

    def set_value(self, newvalue: str):

        try:
            number = float(newvalue)
        except:
            number = math.nan

        self.number = number
        mes = self.format % (self.number,)
        status_list = []
        for set_count in range(int(len(self.range) / 4)):
            status_list.append(self.get_traffic_light_status([self.range[0 + (set_count * 4)],
                                                              self.range[1 + (set_count * 4)],
                                                              self.range[2 + (set_count * 4)],
                                                              self.range[3 + (set_count * 4)]]))

        status = TrafficLight.Red
        for stat in status_list:
            if stat == TrafficLight.Green:
                status = TrafficLight.Green
                break
            if stat == TrafficLight.Orange and status == TrafficLight.Red:
                status = TrafficLight.Orange

        self.text = mes
        self.status = status

    def get_traffic_light_status(self, sub_range: list) -> TrafficLight:
        """
        Check a set of four values within the full range of the yaml file for the status light to apply.

        :param sub_range: a list of four float values
        :return: TrafficLight status
        """

        if np.isnan(self.number) and sub_range[0] == "nan":
            return TrafficLight.Green

        if not isinstance(sub_range[0], str):
            try:
                if ((self.number < sub_range[2]) and (self.number > sub_range[1])):
                    status = TrafficLight.Green
                elif ((self.number < sub_range[3]) and (self.number > sub_range[0])):
                    status = TrafficLight.Orange
                else:
                    status = TrafficLight.Red
            except:
                a = 1
        else:
            status = TrafficLight.Orange
        return status


# ------------------------------------------------------------------------------
#           SensorVariableUI_IncreasingInteger
# ------------------------------------------------------------------------------

class SensorVariableUI_IncreasingInteger(SensorVariableUI_Base):
    """
    Displays integers that should increase (eg counters). Green light is displayed if the value increases between successive
    calls to `set_value`. Red light if the value decreases. If the value stays the same then it is allowed
    """

    def __init__(self, sensor_name: str,
                 type: str,
                 qtapp: ArgolGSE_QtUI_Application,
                 tabiconmanager: Tab_Widget_Manager,
                 cmd_mgr: CMD_Manager,
                 options: Dict[str, Any]):

        super().__init__(sensor_name, type, qtapp, tabiconmanager, cmd_mgr, options)

        self.number: int = -99999  # The current numeric value.
        self.format: str = options.get('format')
        self.num_noincrease_bounds: Tuple[int, int] = options.get('num_noincrease_bounds')
        self.num_nochangeupdates: int = 0

        if self.format is None:
            self.format = "%d"
        if self.num_noincrease_bounds is None:
            self.num_noincrease_bounds = [10000000, 10000001]

        self.status = TrafficLight.Red
        self.text = "--------"

    # ------------------------------------------------------------------------------
    #           set_value(self, newvalue: str, utc: datetime.datetime):
    # ------------------------------------------------------------------------------

    def set_value(self, newvalue: str):

        try:
            number = int(newvalue)
        except:
            number = -99999

        mes = self.format % (number,)
        status = self.set_status(number, self.number)
        self.number = number
        self.text = mes
        self.status = status

    # ------------------------------------------------------------------------------
    #           set_status
    # ------------------------------------------------------------------------------

    def set_status(self, newvalue: int, oldvalue: int) -> TrafficLight:

        status = TrafficLight.Undefined
        if (newvalue > oldvalue):  # if the value has increased since last time
            status = TrafficLight.Green  # then it gets a green light
            self.num_nochangeupdates = 0  # and we reset the update counter
        elif (newvalue < oldvalue):  # if the value has decreased
            status = TrafficLight.Red  # then
        else:
            self.num_nochangeupdates += 1
            if (self.num_nochangeupdates < self.num_noincrease_bounds[0]):
                status = TrafficLight.Green
            elif (self.num_nochangeupdates < self.num_noincrease_bounds[1]):
                status = TrafficLight.Orange
            else:
                status = TrafficLight.Red
            # print(self.num_nochangeupdates)
        return status


# ------------------------------------------------------------------------------
#           SensorVariableUI_IncreasingInteger
# ------------------------------------------------------------------------------

class SensorVariableUI_Integer(SensorVariableUI_Base):
    """
    Displays integers that should increase (eg counters). Green light is displayed if the value increases between successive
    calls to `set_value`. Red light if the value decreases. If the value stays the same then it is allowed
    """

    def __init__(self, sensor_name: str,
                 type: str,
                 qtapp: ArgolGSE_QtUI_Application,
                 tabiconmanager: Tab_Widget_Manager,
                 cmd_mgr: CMD_Manager,
                 options: Dict[str, Any]):

        super().__init__(sensor_name, type, qtapp, tabiconmanager, cmd_mgr, options)

        self.number: int = -99999  # The current numeric value.
        self.format: str = options.get('format')
        self.integer_bounds: Tuple[int, int] = options.get('integer_bounds')

        if self.format is None:
            self.format = "%d"
        self.status = TrafficLight.Red
        self.text = "--------"

    # ------------------------------------------------------------------------------
    #           set_value(self, newvalue: str, utc: datetime.datetime):
    # ------------------------------------------------------------------------------

    def set_value(self, newvalue: str):

        try:
            number = int(newvalue)
        except:
            number = -99999

        mes = self.format % (self.number,)
        status = self.set_status(number, self.number)
        self.number = number
        self.text = mes
        self.status = status

    # ------------------------------------------------------------------------------
    #           set_status
    # ------------------------------------------------------------------------------

    def set_status(self, newvalue: int, oldvalue: int) -> TrafficLight:

        status = TrafficLight.Undefined
        if (self.integer_bounds is None) or ((newvalue >= self.integer_bounds[0]) and (
                newvalue >= self.integer_bounds[0])):  # if the value has increased since last time
            status = TrafficLight.Green  # then it gets a green light
        else:  # if the value has decreased
            status = TrafficLight.Red  # then
        return status


# ------------------------------------------------------------------------------
#           SensorVariableUI_Relay_ButtonMessageForwarder
# ------------------------------------------------------------------------------

class _SensorVariableUI_Relay_ButtonMessageForwarder(QtCore.QObject):

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, parentbutton, func: Callable):
        super().__init__(parentbutton)
        self.pushbutton = parentbutton
        self.func = func
        if (self.pushbutton is not None):
            self.pushbutton.clicked.connect(self.my_button_hasbeen_pressed)

    # ------------------------------------------------------------------------------
    #           my_button_hasbeen_pressed
    # ------------------------------------------------------------------------------

    @QtCore.pyqtSlot(bool)
    def my_button_hasbeen_pressed(self, clicked: bool):
        self.func()
    #


# -----------------------------------------------------------------------------------------------------------------------
#          SensorVariableUI_Relay
# -----------------------------------------------------------------------------------------------------------------------

class SensorVariableUI_Relay(SensorVariableUI_Base):
    """
    A class for rendering Relays and switches on the GSe. This uses a QLabel for status text,
    a QLabel for Icons/pixmap and an additional QPushButton to receive toggle states events.
    The idea is that if the Relay is ON it will be turned OFF. If the relay is OFF it will be turned ON

    The gotcha in this class is that I want to receive Slot messages which means the object needs to
    multiply inherit from both SensorVariableUI_Base and QtCore.QObject. This seems to be too much to ask in Python
    as  both  objects need to be properly initialized and after messing around with super() etc for a few hours I could
    not get it to work. I reworked the design and broke the Slot code off into a separate small class,
    `SensorVariableUI_Relay_ButtonMessageForwarder` that simply forwards the function call into the target class
    """

    def __init__(self, sensorname: str,
                 type: str,
                 qtapp: ArgolGSE_QtUI_Application,
                 tabiconmanager: Tab_Widget_Manager,
                 cmd_mgr: CMD_Manager,
                 options: Dict[str, Any]):

        super().__init__(sensorname, type, qtapp, tabiconmanager, cmd_mgr, options)

        self.command_pending: bool = False
        self.onoffvalue: int = 2
        self.onvals: List[str] = ['TRUE', 'ON', 'YES', 'T', 'Y']
        self.offvals: List[str] = ['FALSE', 'OFF', 'NO', 'F', 'N']
        self.onoffcommands: Tuple[
            str, str] = options.get('onoffcommands')  # 0 = Command string to turn relay to TRUE (when off), 1= Command String to turn relay to FALSE when (On or undefined)
        self.onofflights: Tuple[TrafficLight, TrafficLight] = []
        statuslevel: Tuple[str, str] = options.get('OnOffLights')  #
        if (statuslevel is None):
            statuslevel = ["GREEN", "BLUE"]

        self.onofflights.append(_statusleveltotrafficlight(statuslevel[0]))
        self.onofflights.append(_statusleveltotrafficlight(statuslevel[1]))

        if (self._widget_button is not None):
            self.buttonrelay = _SensorVariableUI_Relay_ButtonMessageForwarder(self._widget_button, self.button_pressed)

    # ------------------------------------------------------------------------------
    #           set_value(self, newvalue: str, utc: datetime.datetime):
    # ------------------------------------------------------------------------------

    def set_value(self, newvalue: str):

        if (self.command_pending):
            onoff = TrafficLight.Orange
            newval = 2
        else:
            v = newvalue.upper()
            on = (v in self.onvals)
            off = (v in self.offvals)

            if (on):
                onoff = self.onofflights[0]
                newval = 0
            elif (off):
                onoff = self.onofflights[1]
                newval = 1
            else:
                onoff = TrafficLight.Orange
                newval = 2

        self.onoffvalue = newval
        self.text = newvalue
        self.status = onoff

    # ------------------------------------------------------------------------------
    #           print_diagnostics
    # ------------------------------------------------------------------------------

    def button_pressed(self):

        self.command_pending = True
        if self.onoffvalue == 1:
            command = self.onoffcommands[0]  # If we are definitely OFF, then turn it ON
        else:
            command = self.onoffcommands[1]  # Otherwise turn it OFF
        print('Sending command ', command)
        self.cmd_manager.send_command(command, self.button_pressed_callback)
        self.status = TrafficLight.Orange

    # ------------------------------------------------------------------------------
    #           button_pressed_callback
    # ------------------------------------------------------------------------------

    def button_pressed_callback(self, success, cmdstr):

        self.command_pending = False


# ------------------------------------------------------------------------------
#           SensorVariableUI_IncreasingInteger
# ------------------------------------------------------------------------------

class SensorVariableUI_MaskedBit(SensorVariableUI_Base):
    """
    Displays integers that should increase (eg counters). Green light is displayed if the value increases between successive
    calls to `set_value`. Red light if the value decreases. If the value stays the same then it is allowed
    """

    def __init__(self, sensor_name: str,
                 type: str,
                 qtapp: ArgolGSE_QtUI_Application,
                 tabiconmanager: Tab_Widget_Manager,
                 cmd_mgr: CMD_Manager,
                 options: Dict[str, Any]):

        super().__init__(sensor_name, type, qtapp, tabiconmanager, cmd_mgr, options)

        self.number: int = 0  # The current numeric value.
        self.incoming_string_base: int = options.get('incoming_string_base')
        self.require_all_maskedbits: bool = options.get('require_all_maskedbits')
        self.bitmask: int = options.get('bitmask')
        self.bitflipmask: int = options.get('bitflipmask')
        self.activehigh: bool = options.get('activehigh')
        self.stat: Tuple[TrafficLight, TrafficLight] = []
        statuslevel: Tuple[str, str] = options.get('OnOffLights')  #

        self.stat.append(_statusleveltotrafficlight(statuslevel[0]))
        self.stat.append(_statusleveltotrafficlight(statuslevel[1]))
        self.text = "---"
        self.status = self.stat[1]

    # ------------------------------------------------------------------------------
    #           set_value(self, newvalue: str, utc: datetime.datetime):
    # ------------------------------------------------------------------------------

    def set_value(self, newvalue: str):

        try:
            if (self.incoming_string_base == 16):
                index = newvalue.find('0x')
                if (index >= 0):
                    newvalue = newvalue[index + 2:]

            number = int(newvalue, self.incoming_string_base)  # Decode the number from the string usinging the proper base
            number = number ^ self.bitflipmask  # Flip value of any bits
            if not self.activehigh:
                number = ~number  # If active when 0 then negate the bits in the number

        except:
            number = int(0)

        if self.require_all_maskedbits:
            isactive = ((number & self.bitmask) == self.bitmask)
        else:
            isactive = ((number & self.bitmask) != 0)

        if isactive:
            stat = self.stat[0]
        else:
            stat = self.stat[1]
        self.number = number
        self.text = newvalue
        self.status = stat


# ------------------------------------------------------------------------------
#           SensorVariableUI_IncreasingInteger
# ------------------------------------------------------------------------------

class SensorVariableUI_String_WithBadValues(SensorVariableUI_Base):
    """
    Displays integers that should increase (eg counters). Green light is displayed if the value increases between successive
    calls to `set_value`. Red light if the value decreases. If the value stays the same then it is allowed
    """

    def __init__(self, sensor_name: str,
                 type: str,
                 qtapp: ArgolGSE_QtUI_Application,
                 tabiconmanager: Tab_Widget_Manager,
                 cmd_mgr: CMD_Manager,
                 options: Dict[str, Any]):

        super().__init__(sensor_name, type, qtapp, tabiconmanager, cmd_mgr, options)

        #     def set_value(self, newvalue: str):
        #       self.text = newvalue
        #       self.status = TrafficLight.Green

        self.badvalues: List[str] = options.get('BadValues')
        self.onofflights: Tuple[TrafficLight, TrafficLight] = []
        statuslevel: Tuple[str, str] = options.get('OnOffLights')  #

        self.onofflights.append(_statusleveltotrafficlight(statuslevel[0]))
        self.onofflights.append(_statusleveltotrafficlight(statuslevel[1]))
        self.status = self.onofflights[1]

    # ------------------------------------------------------------------------------
    #           set_value(self, newvalue: str, utc: datetime.datetime):
    # ------------------------------------------------------------------------------

    def set_value(self, newvalue: str):

        self.text = newvalue
        if ((self.badvalues is None) or (newvalue in self.badvalues)):
            self.status = self.onofflights[1]
        else:
            self.status = self.onofflights[0]
