from typing import Dict, Any, Tuple, List, Union, Callable
from PyQt5 import QtCore, QtGui, QtWidgets
from ..application import ArgolGSE_QtUI_Application
from .trafficlight import TrafficLight


# ------------------------------------------------------------------------------
#           Tab_Widget_Entry
# ------------------------------------------------------------------------------

class Tab_Widget_Entry:
    """
    Stores the status of a single tab within a QTabWidget control. The status is specified
    as a TrafficLight status (the greater the status  the more severe the error) and the index of the tab
    in the tab widget control
    """

    def __init__(self, index: int):
        self.currentstatus: Dict[str, TrafficLight] = {}
        self.tabindex: int = index


# -----------------------------------------------------------------------------------------------------------------------
#           Tab_Widget_Manager
# -----------------------------------------------------------------------------------------------------------------------

class Tab_Widget_Manager:
    """
    A small class that controls the icon displayed in each tab of the GSE tabbed dialog control. The Sensor
    Variables that specify a `tabicon` property in hkvariables.yaml willl automatically report their status
    to the given tab. The tab will report the most severe error of all the icons that are reporting to that tab.

    """

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------

    def __init__(self, widget_name: str, qtapp: ArgolGSE_QtUI_Application):

        self.widget_name: str = widget_name
        self.tab_widget: QtWidgets.QTabWidget = None
        self.tabname_to_index: Dict[str, Tab_Widget_Entry] = {}
        self.g_trafficlights: Dict[TrafficLight, ] = None

        self.tab_widget = qtapp.tab_widget
        for i in range(self.tab_widget.count()):                                                                        # Get the tabs in this tab_widget
            tabName = self.tab_widget.widget(i).objectName()                                                            # Get the name of this tab, as set by in Qt Designer UI,  click on tab, in properties QTabWidget -> currentTabName
            self.tabname_to_index[tabName] = Tab_Widget_Entry(i)                                                        # map the name of this widget to its index position in the tab widget

    # ------------------------------------------------------------------------------
    #           update_tab_status
    # ------------------------------------------------------------------------------

    def update_tab_status(self, tabname: str, field: str, this_status: TrafficLight):

        if (self.tab_widget is not None):
            if (self.g_trafficlights is None):
                self.g_trafficlights = {TrafficLight.Undefined: QtGui.QIcon(":/icons/icons/statusmed_green.png"),
                                        TrafficLight.Green: QtGui.QIcon(":/icons/icons/statusmed_green.png"),           # TrafficLight.Green
                                        TrafficLight.Blue: QtGui.QIcon(":/icons/icons/statusmed_blue.png"),             # TrafficLight.Blue
                                        TrafficLight.Green_Orange: QtGui.QIcon(":/icons/icons/statusmed_orange.png"),   # TrafficLight.Green-Orange
                                        TrafficLight.Orange: QtGui.QIcon(":/icons/icons/statusmed_orange.png"),         # TrafficLight.Orange
                                        TrafficLight.Red: QtGui.QIcon(":/icons/icons/statusmed_red.png")                # TrafficLight.Red
                                        }

            tabentry = self.tabname_to_index.get(tabname)                                                               # Get the ICON entries for this tab
            if (tabentry is not None):                                                                                  # if we have some entries
                newstatus = tabentry.currentstatus.get(field)                                                           # then find the update icons for this area
                if (newstatus is None) or (this_status > newstatus):                                                    # if the status coming in is bigger than what we currently have
                    newstatus = this_status                                                                             # then get the maximum status (most severe)
                tabentry.currentstatus[field] = newstatus                                                               # and setup the new status

                fullstatus = TrafficLight.Undefined                                                                     # start off with the lowest priority setting
                for k, status in tabentry.currentstatus.items():                                                        # now  for each status section on this tab
                    if (status > fullstatus):
                        fullstatus = status                                                                             # find the maximum (most severe) status
                self.tab_widget.setTabIcon(tabentry.tabindex, self.g_trafficlights[fullstatus])                         # and display that status

    # ------------------------------------------------------------------------------
    #           reset_status
    # ------------------------------------------------------------------------------

    def reset_status(self, field: str):

        for k, entry in self.tabname_to_index.items():
            entry.currentstatus[field] = TrafficLight.Undefined
